# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import random
from amd.rocal.plugin.pytorch import ROCALClassificationIterator
from amd.rocal.pipeline import Pipeline
import amd.rocal.fn as fn
import amd.rocal.types as types
import numpy as np
import sys
import os


def draw_patches(img, idx, layout="nchw", dtype="fp32", device="cpu"):
    # image is expected as a tensor
    import cv2
    if device == "cpu":
        image = img.detach().numpy()
    else:
        image = img.cpu().numpy()
    if layout == "nchw":
        image = image.transpose([1, 2, 0])
    if dtype == "fp16":
        image = image.astype("uint8")
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    cv2.imwrite("output_folder/file_reader/" + str(idx) +
                "_" + "train" + ".png", image * 255)


def main():
    if len(sys.argv) < 3:
        print("Please pass image_folder cpu/gpu batch_size")
        exit(0)
    try:
        path = "output_folder/file_reader/"
        isExist = os.path.exists(path)
        if not isExist:
            os.makedirs(path)
    except OSError as error:
        print(error)
    data_path = sys.argv[1]
    rocal_cpu = True if sys.argv[2] == "cpu" else False
    batch_size = int(sys.argv[3])
    num_threads = 1
    device_id = 0
    random_seed = random.SystemRandom().randint(0, 2**32 - 1)

    local_rank = 0
    world_size = 1

    image_classification_train_pipeline = Pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id,
                                                   seed=random_seed, rocal_cpu=rocal_cpu, tensor_layout=types.NHWC, tensor_dtype=types.FLOAT16)

    with image_classification_train_pipeline:
        jpegs, labels = fn.readers.file(file_root=data_path)
        decode = fn.decoders.image_slice(jpegs, output_type=types.RGB,
                                         file_root=data_path, shard_id=local_rank, num_shards=world_size, random_shuffle=True)
        res = fn.resize(decode, resize_width=224, resize_height=224,
                        output_layout=types.NCHW, output_dtype=types.UINT8)
        flip_coin = fn.random.coin_flip(probability=0.5)
        cmnp = fn.crop_mirror_normalize(res,
                                        output_layout=types.NHWC,
                                        output_dtype=types.FLOAT16,
                                        crop=(224, 224),
                                        mirror=flip_coin,
                                        mean=[0.485 * 255, 0.456 *
                                              255, 0.406 * 255],
                                        std=[0.229 * 255, 0.224 * 255, 0.225 * 255])
        image_classification_train_pipeline.set_outputs(cmnp)

    image_classification_train_pipeline.build()
    imageIteratorPipeline = ROCALClassificationIterator(
        image_classification_train_pipeline)
    cnt = 0
    for epoch in range(3):
        print(
            "+++++++++++++++++++++++++++++EPOCH+++++++++++++++++++++++++++++++++++++", epoch)
        for i, it in enumerate(imageIteratorPipeline):
            print(it)
            print(
                "************************************** i *************************************", i)
            for img in it[0]:
                cnt += 1
                draw_patches(img[0], cnt, layout="nhwc",
                             dtype="fp16", device=rocal_cpu)
        imageIteratorPipeline.reset()
    print("*********************************************************************")


if __name__ == "__main__":
    main()



# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import torch
from amd.rocal.pipeline import Pipeline
import amd.rocal.fn as fn
import amd.rocal.types as types
import numpy as np
from parse_config import parse_args


class ROCALVideoIterator(object):
    """
    ROCALVideoIterator for pyTorch.

    Parameters
    ----------
    pipelines : list of amd.rocal.pipeline.Pipeline
                List of pipelines to use
    size : int
           Epoch size.
    """

    def __init__(self, pipelines, tensor_layout=types.NCHW, reverse_channels=False, multiplier=None, offset=None, tensor_dtype=types.FLOAT, display=False, sequence_length=3):

        try:
            assert pipelines is not None, "Number of provided pipelines has to be at least 1"
        except Exception as ex:
            print(ex)

        self.loader = pipelines
        self.tensor_format = tensor_layout
        self.multiplier = multiplier if multiplier else [1.0, 1.0, 1.0]
        self.offset = offset if offset else [0.0, 0.0, 0.0]
        self.reverse_channels = reverse_channels
        self.tensor_dtype = tensor_dtype
        self.batch_size = self.loader._batch_size
        self.rim = self.loader.get_remaining_images()
        self.display = display
        self.iter_num = 0
        self.sequence_length = sequence_length
        print("____________REMAINING IMAGES____________:", self.rim)
        self.output = self.dimensions = self.dtype = None

    def next(self):
        return self.__next__()

    def __next__(self):
        if (self.loader.is_empty()):
            raise StopIteration

        if self.loader.rocal_run() != 0:
            raise StopIteration
        self.output_tensor_list = self.loader.get_output_tensors()
        self.iter_num += 1
        # Copy output from buffer to numpy array
        if self.output is None:
            self.dimensions = self.output_tensor_list[0].dimensions()
            self.dtype = self.output_tensor_list[0].dtype()
            self.layout = self.output_tensor_list[0].layout()
            self.output = np.empty(
                (self.dimensions[0]*self.dimensions[1], self.dimensions[2], self.dimensions[3], self.dimensions[4]), dtype=self.dtype)
        self.output_tensor_list[0].copy_data(self.output)
        img = torch.from_numpy(self.output)
        # Display Frames in a video sequence
        if self.display:
            for batch_i in range(self.batch_size):
                draw_frames(img[batch_i], batch_i, self.iter_num, self.layout)
        return img

    def reset(self):
        self.loader.rocal_reset_loaders()

    def __iter__(self):
        return self

    def __del__(self):
        self.loader.rocal_release()


def draw_frames(img, batch_idx, iter_idx, layout):
    # image is expected as a tensor, bboxes as numpy
    import cv2
    image = img.detach().numpy()
    if layout == 'NFCHW':
        image = image.transpose([1, 2, 0])
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    import os
    if not os.path.exists("output_folder/video_reader"):
        os.makedirs("output_folder/video_reader")
    image = cv2.UMat(image).get()
    cv2.imwrite("output_folder/video_reader/" +
                "iter_"+str(iter_idx)+"_batch_"+str(batch_idx)+".png", image)


def main():
    # Args
    args = parse_args()
    video_path = args.video_path
    rocal_cpu = False if args.rocal_gpu else True
    batch_size = args.batch_size
    user_sequence_length = args.sequence_length
    display = args.display
    num_threads = args.num_threads
    random_seed = args.seed
    tensor_format = types.NFHWC if args.NHWC else types.NFCHW
    tensor_dtype = types.FLOAT16 if args.fp16 else types.FLOAT
    # Create Pipeline instance
    pipe = Pipeline(batch_size=batch_size, num_threads=num_threads, device_id=args.local_rank, seed=random_seed, rocal_cpu=rocal_cpu,
                    tensor_layout=tensor_format, tensor_dtype=tensor_dtype)
    # Use pipeline instance to make calls to reader, decoder & augmentation's
    with pipe:
        images = fn.readers.video(file_root=video_path, sequence_length=user_sequence_length,
                                  random_shuffle=False, image_type=types.RGB)
        crop_size = (512, 960)
        output_images = fn.crop_mirror_normalize(images,
                                                 output_layout=tensor_format,
                                                 output_dtype=tensor_dtype,
                                                 crop=crop_size,
                                                 mean=[0, 0, 0],
                                                 std=[1, 1, 1])
        pipe.set_outputs(output_images)
    # Build the pipeline
    pipe.build()
    # Dataloader
    data_loader = ROCALVideoIterator(pipe, multiplier=pipe._multiplier,
                                     offset=pipe._offset, display=display, sequence_length=user_sequence_length)
    import timeit
    start = timeit.default_timer()
    # Enumerate over the Dataloader
    for epoch in range(int(args.num_epochs)):
        print("EPOCH:::::", epoch)
        for i, it in enumerate(data_loader, 0):
            if args.print_tensor:
                print("**************", i, "*******************")
                print("**************starts*******************")
                print("\n IMAGES : \n", it)
                print("**************ends*******************")
                print("**************", i, "*******************")
        data_loader.reset()
    # Your statements here
    stop = timeit.default_timer()

    print('\n Time: ', stop - start)
    print("##############################  VIDEO READER  SUCCESS  ############################")


if __name__ == '__main__':
    main()



# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##
# @file pipeline.py
#
# @brief File containing the Pipeline class containing the pybind API functions

import rocal_pybind as b
import amd.rocal.types as types
import numpy as np
import ctypes
import functools
import inspect


class Pipeline(object):

    """!Pipeline class internally calls RocalCreate which returns context which will have all
    the info set by the user.

    @param batch_size (int, optional, default = -1)                                                       Batch size of the pipeline. Negative values for this parameter are invalid - the default value may only be used with serialized pipeline (the value stored in serialized pipeline is used instead).
    @param num_threads (int, optional, default = -1)                                                      Number of CPU threads used by the pipeline. Negative values for this parameter are invalid - the default value may only be used with serialized pipeline (the value stored in serialized pipeline is used instead).
    @param device_id (int, optional, default = 0)                                                         Id of GPU used by the pipeline. Negative values for this parameter are invalid
    @param seed (int, optional, default = -1)                                                             Seed used for random number generation. Leaving the default value for this parameter results in random seed.
    @param exec_pipelined (bool, optional, default = True)                                                Whether to execute the pipeline in a way that enables overlapping CPU and GPU computation, typically resultingin faster execution speed, but larger memory consumption.
    @param prefetch_queue_depth (int or {"cpu_size": int, "gpu_size": int}, optional, default = 2)        Depth of the executor pipeline. Deeper pipeline makes ROCAL more resistant to uneven execution time of each batch, but it also consumes more memory for internal buffers. Specifying a dict: ``{ "cpu_size": x, "gpu_size": y }`` instead of an integer will cause the pipeline to use separated queues executor, with buffer queue size `x` for cpu stage and `y` for mixed and gpu stages. It is not supported when both `exec_async` and `exec_pipelined` are set to `False`. Executor will buffer cpu and gpu stages separatelly, and will fill the buffer queues when the first :meth:`amd.rocal.pipeline.Pipeline.run` is issued.
    @param exec_async (bool, optional, default = True)                                                    Whether to execute the pipeline asynchronously. his makes :meth:`amd.rocal.pipeline.Pipeline.run` method run asynchronously with respect to the calling Python thread.
    @param bytes_per_sample  (int, optional, default = 0)                                                 A hint for ROCAL for how much memory to use for its tensors.
    @param rocal_cpu (bool, optional, default = False)                                                    Whether to use CPU or GPU for the pipeline
    @param max_streams (int, optional, default = -1)                                                      Limit the number of HIP streams used by the executor. Value of -1 does not impose a limit. This parameter is currently unused (and behavior of unrestricted number of streams is assumed).
    @param default_cuda_stream_priority (int, optional, default = 0)                                      HIP stream priority used by ROCAL. See `cudaStreamCreateWithPriority` in HIP documentation
    @param tensor_layout (int, optional, default = 0)                                                     Tensor layout used for the augmentations
    @param reverse_channels (int, optional, default = 0)                                                  Whether to reverse channels for the output tensors
    @param mean (int, optional, default = 0)                                                              Mean value used for the image normalization
    @param std (int, optional, default = 0)                                                               Standard deviation value used for the image normalization
    @param tensor_dtype (int, optional, default = 0)                                                      Tensor datatype used for the pipeline
    @param output_memory_type (int, optional, default = 0)                                                Output memory type used for the output tensors
    """
    '''.
    Args: batch_size
          rocal_cpu
          gpu_id (default 0)
          cpu_threads (default 1)
    This returns a context'''
    _handle = None
    _current_pipeline = None

    def __init__(self, batch_size=-1, num_threads=0, device_id=0, seed=1,
                 exec_pipelined=True, prefetch_queue_depth=2,
                 exec_async=True, bytes_per_sample=0,
                 rocal_cpu=False, max_streams=-1, default_cuda_stream_priority=0, tensor_layout=types.NCHW, reverse_channels=False, mean=None, std=None, tensor_dtype=types.FLOAT, output_memory_type=None): 
        if (rocal_cpu):
            self._handle = b.rocalCreate(
                batch_size, types.CPU, device_id, num_threads, prefetch_queue_depth, tensor_dtype)
        else:
            self._handle = b.rocalCreate(
                batch_size, types.GPU, device_id, num_threads, prefetch_queue_depth, tensor_dtype)

        if (b.getStatus(self._handle) == types.OK):
            print("Pipeline has been created succesfully")
        else:
            raise Exception("Failed creating the pipeline")
        self._check_ops = ["CropMirrorNormalize"]
        self._check_crop_ops = ["Resize"]
        self._check_ops_decoder = [
            "ImageDecoder", "ImageDecoderSlice", "ImageDecoderRandomCrop", "ImageDecoderRaw"]
        self._check_ops_reader = ["labelReader", "TFRecordReaderClassification", "TFRecordReaderDetection",
                                  "COCOReader", "Caffe2Reader", "Caffe2ReaderDetection", "CaffeReader", "CaffeReaderDetection", "NumpyReader"]
        self._batch_size = batch_size
        self._num_threads = num_threads
        self._device_id = device_id
        self._output_memory_type = output_memory_type if output_memory_type else (
            types.HOST_MEMORY if rocal_cpu else types.DEVICE_MEMORY)
        self._seed = seed
        self._exec_pipelined = exec_pipelined
        self._prefetch_queue_depth = prefetch_queue_depth
        self._exec_async = exec_async
        self._bytes_per_sample = bytes_per_sample
        self._rocal_cpu = rocal_cpu
        self._max_streams = max_streams
        self._default_cuda_stream_priority = default_cuda_stream_priority
        self._tensor_layout = tensor_layout
        self._tensor_dtype = tensor_dtype
        self._multiplier = list(map(lambda x: 1/x, std)
                                ) if std else [1.0, 1.0, 1.0]
        self._offset = list(map(lambda x, y: -(x/y), mean, std)
                            ) if mean and std else [0.0, 0.0, 0.0]
        self._reverse_channels = reverse_channels
        self._img_h = None
        self._img_w = None
        self._shuffle = None
        self._name = None
        self._anchors = None
        self._box_encoder = None
        self._box_iou_matcher = None
        self._encode_tensor = None
        self._num_classes = None
        self._one_hot_encoding = False
        self._castLabels = False
        self._current_pipeline = None
        self._reader = None
        self._define_graph_set = False
        self.set_seed(self._seed)
        self._is_external_source_operator = False
        self._external_source = None
        self._external_source_mode = None
        self._last_batch_policy = None

    def build(self):
        """!Build the pipeline using rocalVerify call
        """
        status = b.rocalVerify(self._handle)
        if (status != types.OK):
            print("Verify graph failed")
            exit(0)
        return self

    def rocal_run(self):
        """! Run the pipeline using rocalRun call
        """
        status = b.rocalRun(self._handle)
        return status

    def define_graph(self):
        """!This function is defined by the user to construct the
        graph of operations for their pipeline.
        It returns a list of outputs created by calling ROCAL Operators."""
        print("define_graph is deprecated")
        raise NotImplementedError

    def get_handle(self):
        return self._handle

    def copyToExternalTensor(self, array,  multiplier, offset, reverse_channels, tensor_format, tensor_dtype, max_roi_height=0, max_roi_width=0):
        b.rocalToTensor(self._handle, ctypes.c_void_p(array.data_ptr()), tensor_format, tensor_dtype,
                        multiplier[0], multiplier[1], multiplier[2], offset[0], offset[1], offset[2], (1 if reverse_channels else 0), self._output_memory_type, max_roi_height, max_roi_width)

    def get_one_hot_encoded_labels(self, array_ptr, dest_device_type):
            b.getOneHotEncodedLabels(self._handle, array_ptr, self._num_classes, dest_device_type)

    def set_outputs(self, *output_list):
        b.setOutputs(self._handle, len(output_list), output_list)

    def __enter__(self):
        Pipeline._current_pipeline = self
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass

    def set_seed(self, seed=0):
        return b.setSeed(seed)

    @classmethod
    def create_int_param(self, value=1):
        return b.createIntParameter(value)

    @classmethod
    def create_float_param(self, value=1):
        return b.createFloatParameter(value)

    @classmethod
    def update_int_param(self, value=1, param=1):
        b.updateIntParameter(value, param)

    @classmethod
    def update_float_param(self, value=1, param=1):
        b.updateFloatParameter(value, param)

    @classmethod
    def get_int_value(self, param):
        return b.getIntValue(param)

    @classmethod
    def get_float_value(self, param):
        return b.getFloatValue(param)

    def get_image_name(self, array_len):
        return b.getImageName(self._handle, array_len)

    def get_image_id(self, array):
        b.getImageId(self._handle, array)

    def get_bounding_box_count(self):
        return b.getBoundingBoxCount(self._handle)

    def get_bounding_box_labels(self):
        return b.getBoundingBoxLabels(self._handle)

    def get_bounding_box_cords(self):
        return b.getBoundingBoxCords(self._handle)
    
    def get_ascii_datas(self):
        return b.getAsciiDatas(self._handle)

    def get_mask_count(self, array):
        return b.getMaskCount(self._handle, array)

    def get_mask_coordinates(self, array_count, array):
        return b.getMaskCoordinates(self._handle, array_count, array)
    
    def get_image_labels(self):
        return b.getImageLabels(self._handle)

    def copy_encoded_boxes_and_lables(self, bbox_array, label_array):
        b.rocalCopyEncodedBoxesAndLables(self._handle, bbox_array, label_array)

    def get_encoded_boxes_and_lables(self, batch_size, num_anchors):
        return b.rocalGetEncodedBoxesAndLables(self._handle, batch_size, num_anchors)

    def get_img_sizes(self, array):
        return b.getImgSizes(self._handle, array)

    def get_roi_img_sizes(self, array):
        return b.getROIImgSizes(self._handle, array)
    
    def get_image_name_length(self, idx):
        return b.getImageNameLen(self._handle, idx)

    def get_remaining_images(self):
        return b.getRemainingImages(self._handle)

    def rocal_release(self):
        return b.rocalRelease(self._handle)

    def rocal_reset_loaders(self):
        return b.rocalResetLoaders(self._handle)

    def is_empty(self):
        return b.isEmpty(self._handle)

    def timing_info(self):
        return b.getTimingInfo(self._handle)

    def get_matched_indices(self):
        return b.getMatchedIndices(self._handle)

    def get_output_tensors(self):
        return b.getOutputTensors(self._handle)
    
    def get_last_batch_padded_size(self):
        return b.getLastBatchPaddedSize(self._handle)

    def run(self):
        """
        It raises StopIteration if data set reached its end.
        return:
        :return:
        A list of `rocalTensorList` objects for respective pipeline outputs.
        """
        try:
            if self.get_remaining_images() > 0:
                self.rocal_run()
                return b.getOutputTensors(self._handle)
        except:
            raise StopIteration


def _discriminate_args(func, **func_kwargs):
    """!Split args on those applicable to Pipeline constructor and the decorated function."""
    func_argspec = inspect.getfullargspec(func)
    ctor_argspec = inspect.getfullargspec(Pipeline.__init__)

    if 'debug' not in func_argspec.args and 'debug' not in func_argspec.kwonlyargs:
        func_kwargs.pop('debug', False)

    ctor_args = {}
    fn_args = {}

    if func_argspec.varkw is not None:
        raise TypeError(
            f"Using variadic keyword argument `**{func_argspec.varkw}` in a  "
            f"graph-defining function is not allowed.")

    for farg in func_kwargs.items():
        is_ctor_arg = farg[0] in ctor_argspec.args or farg[0] in ctor_argspec.kwonlyargs
        is_fn_arg = farg[0] in func_argspec.args or farg[0] in func_argspec.kwonlyargs
        if is_fn_arg:
            fn_args[farg[0]] = farg[1]
            if is_ctor_arg:
                print(
                    "Warning: the argument `{farg[0]}` shadows a Pipeline constructor "
                    "argument of the same name.")
        elif is_ctor_arg:
            ctor_args[farg[0]] = farg[1]
        else:
            assert False, f"This shouldn't happen. Please double-check the `{farg[0]}` argument"

    return ctor_args, fn_args


def pipeline_def(fn=None, **pipeline_kwargs):
    """!
    Decorator that converts a graph definition function into a rocAL pipeline factory.

    A graph definition function is a function that returns intended pipeline outputs.
    You can decorate this function with ``@pipeline_def``::

        @pipeline_def
        def my_pipe(flip_vertical, flip_horizontal):
            ''' Creates a rocAL pipeline, which returns flipped and original images '''
            data, _ = fn.readers.file(file_root=images_dir)
            img = fn.decoders.image(data, device="mixed")
            flipped = fn.flip(img, horizontal=flip_horizontal, vertical=flip_vertical)
            return flipped, img

    The decorated function returns a rocAL Pipeline object::

        pipe = my_pipe(True, False)
        # pipe.build()  # the pipeline is not configured properly yet

    A pipeline requires additional parameters such as batch size, number of worker threads,
    GPU device id and so on (see :meth:`amd.rocal.Pipeline()` for a
    complete list of pipeline parameters).
    These parameters can be supplied as additional keyword arguments,
    passed to the decorated function::

        pipe = my_pipe(True, False, batch_size=32, num_threads=1, device_id=0)
        pipe.build()  # the pipeline is properly configured, we can build it now

    The outputs from the original function became the outputs of the Pipeline::

        flipped, img = pipe.run()

    When some of the pipeline parameters are fixed, they can be specified by name in the decorator::

        @pipeline_def(batch_size=42, num_threads=3)
        def my_pipe(flip_vertical, flip_horizontal):
            ...

    Any Pipeline constructor parameter passed later when calling the decorated function will
    override the decorator-defined params::

        @pipeline_def(batch_size=32, num_threads=3)
        def my_pipe():
            data = fn.external_source(source=my_generator)
            return data

        pipe = my_pipe(batch_size=128)  # batch_size=128 overrides batch_size=32

    .. warning::

        The arguments of the function being decorated can shadow pipeline constructor arguments -
        in which case there's no way to alter their values.

    .. note::

        Using ``**kwargs`` (variadic keyword arguments) in graph-defining function is not allowed.
        They may result in unwanted, silent hijacking of some arguments of the same name by
        Pipeline constructor. Code written this way would cease to work with future versions of rocAL
        when new parameters are added to the Pipeline constructor.

    To access any pipeline arguments within the body of a ``@pipeline_def`` function, the function
    :meth:`amd.rocal.Pipeline.current()` can be used:: ( note: this is not supported yet)

        @pipeline_def()
        def my_pipe():
            pipe = Pipeline.current()
            batch_size = pipe.batch_size
            num_threads = pipe.num_threads
            ...

        pipe = my_pipe(batch_size=42, num_threads=3)
        ...
    """

    def actual_decorator(func):

        @functools.wraps(func)
        def create_pipeline(*args, **kwargs):
            ctor_args, fn_kwargs = _discriminate_args(func, **kwargs)
            # Merge and overwrite dict
            pipe = Pipeline(**{**pipeline_kwargs, **ctor_args})
            with pipe:
                pipe_outputs = func(*args, **fn_kwargs)
                if isinstance(pipe_outputs, tuple):
                    po = pipe_outputs
                elif pipe_outputs is None:
                    po = ()
                else:
                    po = (pipe_outputs, )
                pipe.set_outputs(*po)
            return pipe

        # Add `is_pipeline_def` attribute to the function marked as `@pipeline_def`
        create_pipeline._is_pipeline_def = True
        return create_pipeline

    return actual_decorator(fn) if fn else actual_decorator



# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from argparse import ArgumentParser
import random


def parse_args():
    parser = ArgumentParser(description="Train Single Shot MultiBox Detector"
                                        " on COCO")

    common_group = parser.add_argument_group(
        'common', 'common-related options')
    # Data-related
    common_group.add_argument('--image-dataset-path', '-d', type=str,
                              help='image folder files')
    common_group.add_argument('--batch-size', '-b', type=int, default=10,
                              help='number of examples for each iteration')
    common_group.add_argument('--display', action="store_true",
                              help='--display:to display output from the pipeline')
    common_group.add_argument('--no-display', dest='display', action="store_false",
                              help='--no-display:to not display output from the pipeline')
    # case when none of the above is specified
    parser.set_defaults(display=False)
    common_group.add_argument('--max-height', '-mh', type=int, default=1000,
                              help='maximum height set during decoding')
    common_group.add_argument('--max-width', '-mw', type=int, default=1000,
                              help='maximum width set during decoding')
    common_group.add_argument('--color-format', '-c', type=int, default=1,
                              help='color format used during decoding')

    common_group.add_argument('--print_tensor', action="store_true",
                              help='--print_tensor: to print tensor output from the pipeline')
    common_group.add_argument('--no-print_tensor', dest='print_tensor', action="store_false",
                              help='--no-print_tensor: to not print tensor output from the pipeline')
    # case when none of the above is specified
    parser.set_defaults(print_tensor=False)

    common_group.add_argument('--classification', action="store_true",
                              help='--classification: to use for classification')
    common_group.add_argument('--no-classification', dest='classification', action="store_false",
                              help='--no-classification: to use for detection pipeline')
    # case when none of the above is specified
    parser.set_defaults(classification=True)

    common_group.add_argument('--rocal-gpu', default=False, action="store_true",
                              help='--use_gpu to use gpu')
    common_group.add_argument('--no-rocal-gpu', dest='rocal-gpu', action="store_false",
                              help='--no-rocal-gpu to use cpu backend')
    
    common_group.add_argument('--one-hot-encode', action="store_true",
                              help='--one-hot-encode: to use for one-hot-encoding of labels')
    common_group.add_argument('--no-one-hot-encode', dest='one-hot-encode', action="store_false",
                              help='--no-one-hot-encode: to used when we do not want to one hot encode the labels')
    # case when none of the above is specified
    parser.set_defaults(one_hot_encode=False)

    common_group.add_argument('--NHWC', action='store_true',
                              help='run input pipeline NHWC format')
    common_group.add_argument('--no-NHWC', dest='NHWC', action='store_false',
                              help='run input pipeline NCHW format')
    parser.set_defaults(NHWC=True)  # case when none of the above is specified

    common_group.add_argument('--fp16', default=False, action='store_true',
                              help='run input pipeline fp16 format')

    common_group.add_argument('--local-rank', type=int, default=0,
                              help='Device ID used by rocAL pipeline')
    common_group.add_argument('--world-size', '-w', type=int, default=1,
                              help='number of partitions to split the dataset')
    common_group.add_argument('--num-threads', '-nt', type=int, default=1,
                              help='number of CPU threads used by the rocAL pipeline.')
    common_group.add_argument('--num-epochs', '-e', type=int, default=1,
                              help='number of epochs to run')
    common_group.add_argument('--seed', '-s', type=int, default=random.SystemRandom().randint(0, 2**32 - 1),
                              help='manually set random seed')

    # unit_test.py related options
    python_unit_test = parser.add_argument_group(
        'python-unittest', 'python-unittest-related options')
    python_unit_test.add_argument('--reader-type', '-r', type=str, default="file",
                                  help='Reader used for reading and decoding the images')
    python_unit_test.add_argument('--augmentation-name', '-aug_name', type=str, default="resize",
                                  help='refer python unit test for all augmentation names ')
    python_unit_test.add_argument('--output-file-name', '-f', type=str, default="",
                                  help='file name to save the augmentation outputs')
    python_unit_test.add_argument('--interpolation-type', '-i', type=int, default=1,
                                  help='interpolation type used for resize and crop')
    python_unit_test.add_argument('--scaling-mode', '-sm', type=int, default=0,
                                  help='scaling mode type used for resize')

    # audio_unittests.py related options
    audio_unit_test = parser.add_argument_group(
        'audio-python-unittest', 'audio-python-unittest-related options')
    audio_unit_test.add_argument('--audio_path', type=str, default="",
                                  help='audio files path')
    audio_unit_test.add_argument('--file_list_path', type=str, default="",
                                  help='file list path')
    audio_unit_test.add_argument('--test_case', type=int, default=None,
                                  help='test case')
    audio_unit_test.add_argument('--qa_mode', type=int, default=1,
                                  help='enable qa mode to compare audio output with ref outputs')
    # coco_reader.py related options
    coco_reader = parser.add_argument_group(
        'coco-pipeline', 'coco-pipeline-related options')
    coco_reader.add_argument('--json-path', '-json-path', type=str,
                               help='coco dataset json path')
    # caffe_reader.py related options
    caffe_pipeline = parser.add_argument_group(
        'caffe-pipeline', 'caffe-pipeline-related options')
    caffe_pipeline.add_argument('--detection', '-detection', type=str,
                                help='detection')
    # video_pipeline.py related options
    video_pipeline = parser.add_argument_group(
        'video-pipeline', 'video-pipeline-related options')
    video_pipeline.add_argument('--video-path', '-video-path', type=str,
                                help='video path')
    video_pipeline.add_argument('--sequence-length', '-sequence-length', type=int,
                                help='video path')
    # web_dataset_reader.py related options
    web_dataset_reader = parser.add_argument_group(
        'webdataset-pipeline', 'webdataset-pipeline-related options')
    web_dataset_reader.add_argument('--index-path', '-index-path', type=str,
                               help='web dataset index path')

    return parser.parse_args()



# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##
# @file random.py
#
# @brief File containing randomization functions used for creating RNG generators

import rocal_pybind as b
from amd.rocal.pipeline import Pipeline


def coin_flip(*inputs, probability=0.5):
    """!coin flip with a given probability of success.

        @param inputs         list of input arguments.
        @param probability    Probability of getting a "success" outcome.

        @return    An array of coin flip outcomes.
    """
    values = [0, 1]
    frequencies = [1 - probability, probability]
    output_array = b.createIntRand(values, frequencies)
    return output_array

def uniform(*inputs, range=[-1.0, 1.0]):
    """
    Creates a rocalTensor from an uniform distribution with given range.
    """
    kwargs_pybind = { "inputs": inputs[0], "is_output": False, "range": range }
    output_uniform_distribution = b.uniformDistribution(Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    return (output_uniform_distribution)

def normal(*inputs, mean=0.0, stddev=1.0):
    """
    Creates a rocalTensor from an normal distribution with given mean and stddev.
    """
    kwargs_pybind = { "inputs": inputs[0], "is_output": False, "mean": mean, "stddev": stddev }
    output_normal_distribution = b.normalDistribution(Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    return (output_normal_distribution)



# Copyright (c) 2024 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from amd.rocal.pipeline import pipeline_def
from amd.rocal.plugin.pytorch import ROCALAudioIterator
import amd.rocal.fn as fn
import amd.rocal.types as types
import random
import os
import sys
import matplotlib.pyplot as plt
import torch
import numpy as np
import timeit
from parse_config import parse_args

np.set_printoptions(threshold=1000, edgeitems=10000)
seed = random.SystemRandom().randint(0, 2**32 - 1)

test_case_augmentation_map = {
    0: "audio_decoder",
    1: "preemphasis_filter",
    2: "spectrogram",
    3: "downmix",
    4: "to_decibels",
    5: "resample",
    6: "tensor_add_tensor",
    7: "tensor_mul_scalar",
    8: "non_silent_region",
    9: "slice",
    10: "mel_filter_bank",
    11: "normalize"
}

def plot_audio_wav(audio_tensor, idx):
    # audio is expected as a tensor
    audio_data = audio_tensor.detach().numpy()
    audio_data = audio_data.flatten()
    plt.plot(audio_data)
    plt.savefig("output_folder/audio_reader/" + str(idx) + ".png")
    plt.close()

def verify_non_silent_region_output(output_list, rocal_data_path, test_results, case_name):
    ref_path = f'{rocal_data_path}/rocal_data/GoldenOutputsTensor/reference_outputs_audio/{case_name}_output.bin'
    data_array = np.fromfile(ref_path, dtype=np.int32)
    begin_nsr = output_list[0].detach().numpy()
    length_nsr = output_list[1].detach().numpy()

    if begin_nsr[0] == data_array[0] and length_nsr == data_array[1]:
        print("PASSED!")
        test_results[case_name] = "PASSED"
    else:
        print("FAILED!")
        test_results[case_name] = "FAILED"

def verify_output(audio_tensor, rocal_data_path, roi_tensor, test_results, case_name, dimensions):
    ref_path = f'{rocal_data_path}/rocal_data/GoldenOutputsTensor/reference_outputs_audio/{case_name}_output.bin'
    data_array = np.fromfile(ref_path, dtype=np.float32)
    audio_data = audio_tensor[0].detach().numpy().flatten()
    roi_data = roi_tensor.detach().numpy()
    buffer_size = roi_data[0] * roi_data[1]
    matched_indices = 0
    atol = 1e-20 if case_name != "normalize" else 1e-5  # Reducing absolute tolerance for normalize test case to fix test case failures
    for i in range(roi_data[0]):
        for j in range(roi_data[1]):
            ref_val = data_array[i * roi_data[1] + j]
            out_val = audio_data[i * roi_data[1] + j]   # Stride upto max_roi
            # ensuring that out_val is not exactly zero while ref_val is non-zero.
            invalid_comparison = (out_val == 0.0) and (ref_val != 0.0)
            #comparing the absolute difference between the output value (out_val) and the reference value (ref_val) with a tolerance threshold.
            if not invalid_comparison and abs(out_val - ref_val) < atol:
                matched_indices += 1

    # Print results
    print(f"Results for {case_name}:")
    if matched_indices == buffer_size and matched_indices != 0:
        print("PASSED!")
        test_results[case_name] = "PASSED"
    else:
        print("FAILED!")
        test_results[case_name] = "FAILED"

@pipeline_def(seed=seed)
def audio_decoder_pipeline(path, file_list, downmix=False):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    return fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=downmix,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)

@pipeline_def(seed=seed)
def pre_emphasis_filter_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    return fn.preemphasis_filter(decoded_audio)

@pipeline_def(seed=seed)
def spectrogram_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    spec = fn.spectrogram(
        decoded_audio,
        nfft=512,
        window_length=320,
        window_step=160,
        output_dtype = types.FLOAT)
    return spec

@pipeline_def(seed=seed)
def to_decibels_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    return fn.to_decibels(
            decoded_audio,
            multiplier=np.log(10),
            reference=1.0,
            cutoff_db=np.log(1e-20),
            output_dtype=types.FLOAT)

@pipeline_def(seed=seed)
def resample_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=True,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    input_sample_rate = 16000.00
    uniform_distribution_resample = fn.random.uniform(decoded_audio, range=[1.15, 1.15])
    resampled_rate = uniform_distribution_resample * input_sample_rate
    return fn.resample(
        decoded_audio,
        resample_rate=resampled_rate,
        resample_hint=1.15 * 255840,
        output_datatype=types.FLOAT)

@pipeline_def(seed=seed)
def tensor_add_tensor_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    uniform_distribution_sample = fn.random.uniform(decoded_audio, range=[1.15, 1.15])
    return decoded_audio + uniform_distribution_sample

@pipeline_def(seed=seed)
def tensor_mul_scalar_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=True,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    return decoded_audio * 1.15

@pipeline_def(seed=seed)
def non_silent_region(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    begin, length = fn.nonsilent_region(decoded_audio, cutoff_db=-60)
    return begin, length

@pipeline_def(seed=seed)
def non_silent_region_and_slice(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    begin, length = fn.nonsilent_region(decoded_audio, cutoff_db=-60)
    trim_silence = fn.slice(
        decoded_audio,
        anchor=[begin],
        shape=[length],
        rocal_tensor_output_type = types.FLOAT)
    return trim_silence

@pipeline_def(seed=seed)
def mel_filter_bank_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    spec = fn.spectrogram(
        decoded_audio,
        nfft=512,
        window_length=320,
        window_step=160,
        output_dtype=types.FLOAT)
    return fn.mel_filter_bank(
        spec,
        freq_high=8000,
        freq_low=0.0,
        nfilter=80,
        sample_rate=16000)

@pipeline_def(seed=seed)
def normalize_pipeline(path, file_list):
    audio, labels = fn.readers.file(file_root=path, file_list=file_list)
    decoded_audio = fn.decoders.audio(
        audio,
        file_root=path,
        file_list_path=file_list,
        downmix=False,
        shard_id=0,
        num_shards=1,
        stick_to_shard=False,
        last_batch_policy=types.LAST_BATCH_DROP, pad_last_batch_repeated=False)
    spec = fn.spectrogram(
        decoded_audio,
        nfft=512,
        window_length=320,
        window_step=160,
        output_dtype=types.FLOAT)
    return fn.normalize(
        spec,
        axes=[1])

def main():
    args = parse_args()

    audio_path = args.audio_path
    file_list = args.file_list_path
    rocal_cpu = False if args.rocal_gpu else True
    batch_size = args.batch_size
    test_case = args.test_case
    qa_mode = args.qa_mode
    num_threads = 1
    device_id = 0
    rocal_data_path = os.environ.get("ROCAL_DATA_PATH")

    case_list = list(test_case_augmentation_map.keys())

    if test_case is not None: 
        if test_case not in case_list:
            print(" Invalid Test Case! ")
            exit()
        else:
            case_list = [test_case]

    if args.display:
        try:
            path = "output_folder/audio_reader"
            isExist = os.path.exists(path)
            if not isExist:
                os.makedirs(path)
        except OSError as error:
            print(error)

    if rocal_data_path is None:
        print("Need to export ROCAL_DATA_PATH")
        sys.exit()
    if not rocal_cpu:
        print("The GPU support for Audio is not given yet. Running on CPU")
        rocal_cpu = True
    if not audio_path and not file_list:
        audio_path = f'{rocal_data_path}/rocal_data/audio/'
        file_list = f'{rocal_data_path}/rocal_data/audio/wav_file_list.txt'
        downmix_audio_path = f'{rocal_data_path}/rocal_data/multi_channel_wav/'
    else:
        print("QA mode is disabled for custom audio data")
        qa_mode = 0
    if qa_mode and batch_size != 1:
        print("QA mode is enabled. Batch size is set to 1.")
        batch_size = 1

    print("*********************************************************************")
    test_results = {}
    for case in case_list:
        case_name = test_case_augmentation_map.get(case)
        if case_name == "audio_decoder":
            audio_pipeline = audio_decoder_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "preemphasis_filter":
            audio_pipeline = pre_emphasis_filter_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "spectrogram":
            audio_pipeline = spectrogram_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "downmix":
            audio_pipeline = audio_decoder_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu,
                                                    path=downmix_audio_path if qa_mode else audio_path, file_list="" if qa_mode else file_list, downmix=True)
        if case_name == "to_decibels":
            audio_pipeline = to_decibels_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "slice":
            audio_pipeline = non_silent_region_and_slice(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "non_silent_region":
            audio_pipeline = non_silent_region(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "resample":
            audio_pipeline = resample_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "tensor_add_tensor":
            audio_pipeline = tensor_add_tensor_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "tensor_mul_scalar":
            audio_pipeline = tensor_mul_scalar_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "mel_filter_bank":
            audio_pipeline = mel_filter_bank_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        if case_name == "normalize":
            audio_pipeline = normalize_pipeline(batch_size=batch_size, num_threads=num_threads, device_id=device_id, rocal_cpu=rocal_cpu, path=audio_path, file_list=file_list)
        audio_pipeline.build()
        audio_loader = ROCALAudioIterator(audio_pipeline, auto_reset=True)
        output_tensor_list = audio_pipeline.get_output_tensors()
        dimensions = output_tensor_list[0].dimensions()
        cnt = 0
        start = timeit.default_timer()
        # Enumerate over the Dataloader
        for e in range(int(args.num_epochs)):
            print("Epoch :: ", e)
            torch.set_printoptions(threshold=5000, profile="full", edgeitems=100)
            for i, output_list in enumerate(audio_loader):
                for x in range(len(output_list[0])):
                    for audio_tensor, label, roi in zip(output_list[0][x], output_list[1], output_list[2]):
                        if args.print_tensor:
                            print("label", label)
                            print("Audio", audio_tensor)
                            print("Roi", roi)
                        if args.display:
                            plot_audio_wav(audio_tensor, cnt)
                        cnt+=1
            if qa_mode :
                if case_name == "non_silent_region":
                    verify_non_silent_region_output(output_list[0], rocal_data_path, test_results, case_name)
                else:
                    verify_output(output_list[0], rocal_data_path, roi, test_results, case_name, dimensions)
            print("EPOCH DONE", e)

        stop = timeit.default_timer()
        print('\nTime: ', stop - start)

    if qa_mode:
        passed_cases = []
        failed_cases = []

        for augmentation_name, result in test_results.items():
            if result == "PASSED":
                passed_cases.append(augmentation_name)
            else:
                failed_cases.append(augmentation_name)

        print("Number of PASSED tests:", len(passed_cases))
        print(passed_cases)
        print("Number of FAILED tests:", len(failed_cases))
        print(failed_cases)


if __name__ == "__main__":
    main()



# Copyright (c) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##
# @file jax.py
#
# @brief File containing iterators and functions for JAX framework

import jax
import jax.dlpack
import jax.numpy as jnp
from jax.sharding import NamedSharding, PositionalSharding

try:
    from clu.data.dataset_iterator import ArraySpec
    CLU_FOUND = True
except ImportError:
    CLU_FOUND = False

import threading
import concurrent.futures
import numpy as np
from packaging.version import Version

import rocal_pybind as b
import amd.rocal.types as types

if Version(jax.__version__) < Version("0.4.23"):
    print('rocAL only supports jax versions >= 0.4.23')


def convert_to_jax_array(array):
    """Converts input DLPack tensor to JAX array.

    Args:
        array (DLPack tensor):
            array to be converted to JAX array.

    Returns:
        jax.Array: JAX array with the same values and device as input array.
    """
    jax_array = jax.dlpack.from_dlpack(array, copy=True)
    return jax_array

class ROCALJaxIterator(object):
    """Initializes the ROCAL JAX iterator.

    Args:
        pipelines (list of Pipeline objects): List of rocAL pipelines to use.
        sharding (JAX sharding, optional): JAX sharding to use for placing outputs on devices. Defaults to None.
    """

    def __init__(self, pipelines, sharding=None):
        if not isinstance(pipelines, list):
            pipelines = [pipelines]
        self.pipelines = pipelines
        self.num_devices = len(pipelines)
        self.batch_size = pipelines[0]._batch_size

        self.iterator_length = b.getRemainingImages(
            pipelines[0]._handle) // self.batch_size  # Length should be the same across all pipelines
        self.last_batch_policy = pipelines[0]._last_batch_policy
        assert (
            self.last_batch_policy != types.LAST_BATCH_PARTIAL
        ), "JAX iterator does not support partial last batch policy."

        if sharding is not None:
            assert isinstance(
                sharding, (NamedSharding, PositionalSharding)
            ), "`sharding` should be an instance of `NamedSharding` or `PositionalSharding`"
        self.sharding = sharding
        self._has_started = False

    def next(self):
        """Returns the next batch of data."""
        return self.__next__()

    def __next__(self):
        """Returns the next batch of data."""
        self._has_started = True
        pipeline_outputs = []
        for pipeline in self.pipelines:
            if pipeline.rocal_run() != 0:
                raise StopIteration
            output_tensor_list = pipeline.get_output_tensors()

            output_list = []
            device_id = pipeline._device_id
            if pipeline._name is None:
                pipeline._name = pipeline._reader
            labels_size = ((self.batch_size * pipeline._num_classes)
                                if pipeline._one_hot_encoding else self.batch_size)
            for i in range(len(output_tensor_list)):
                output = convert_to_jax_array(
                    output_tensor_list[i].__dlpack__(device_id))
                output_list.append(output)

            if pipeline._name == "labelReader":
                if pipeline._one_hot_encoding:
                    labels = np.empty(labels_size, dtype="int32")
                    pipeline.get_one_hot_encoded_labels(
                        labels.ctypes.data, pipeline._output_memory_type)
                    labels_tensor = labels.reshape(
                        -1, self.batch_size, pipeline._num_classes)
                    labels_tensor = convert_to_jax_array(
                        labels_tensor)
                else:
                    labels = pipeline.get_image_labels()
                    labels_tensor = labels.astype(dtype=np.int32)
                    labels_tensor = convert_to_jax_array(
                        labels_tensor)
                labels_tensor = jax.device_put(
                    labels_tensor, output_list[0].device)
                output_list.append(labels_tensor)
            pipeline_outputs.append(output_list)

        if self.num_devices == 1 and self.sharding is None:
            return pipeline_outputs[0]

        sharded_outputs = []
        for i in range(len(pipeline_outputs[0])):
            individual_outputs = []
            for pipeline_id in range(self.num_devices):
                individual_outputs.append(pipeline_outputs[pipeline_id][i])
            for output in individual_outputs:
                assert output.shape == individual_outputs[0].shape, "All outputs should have the same shape"
            if self.sharding is not None:
                sharded_outputs.append(self.place_output_with_sharding(
                    individual_outputs))
            else:
                sharded_outputs.append(self.place_output_with_device_put(
                    individual_outputs))
        return sharded_outputs

    def reset(self):
        """Resets the iterator for the next epoch."""
        for pipeline in self.pipelines:
            b.rocalResetLoaders(pipeline._handle)

    def place_output_with_device_put(self, individual_outputs):
        """Builds sharded jax.Array with `jax.device_put_sharded` - compatible
        with pmapped JAX functions.
        """
        output_devices = tuple(
            map(lambda jax_shard: jax_shard.device, individual_outputs)
        )

        if len(output_devices) != len(set(output_devices)):
            if len(set(output_devices)) != 1:
                raise AssertionError(
                    (
                        "JAX iterator requires shards to be placed on different devices "
                        "or all on the same device."
                    )
                )
            else:
                # All shards are on one device (CPU or one GPU)
                return jnp.stack(individual_outputs)
        else:
            return jax.device_put_sharded(individual_outputs, output_devices)

    def place_output_with_sharding(self, individual_outputs):
        """Builds sharded jax.Array with `jax.make_array_from_single_device_arrays`-
        compatible with automatic parallelization with JAX.
        """
        shard_shape = individual_outputs[0].shape
        data_rank = len(shard_shape)
        num_devices = len(individual_outputs)

        # Calculate the global shape. The first (batch) dimension is aggregated.
        global_shape = (num_devices * shard_shape[0],) + shard_shape[1:]

        if isinstance(self.sharding, NamedSharding):
            # Assumes the user wants to shard along the first mesh axis
            # and replicate across the other data dimensions.
            mesh = self.sharding.mesh
            # Create a PartitionSpec matching the data's rank.
            # e.g., for 4D data and mesh axis 'data', this becomes P('data', None, None, None)
            partition_spec = jax.sharding.PartitionSpec(
                mesh.axis_names[0], *(None for _ in range(data_rank - 1)))
            compatible_sharding = NamedSharding(mesh, partition_spec)
        else:
            # Assumes the user wants to shard along the single device axis
            # and replicate across the other data dimensions.
            devices = self.sharding._devices
            # Create a sharding shape matching the data's rank.
            # e.g., for 4D data and 8 devices, this becomes (8, 1, 1, 1)
            sharding_spec_shape = (num_devices,) + (1,) * (data_rank - 1)
            compatible_sharding = PositionalSharding(devices).reshape(sharding_spec_shape)

        return jax.make_array_from_single_device_arrays(
            global_shape, compatible_sharding, individual_outputs
        )

    def __iter__(self):
        """Returns the iterator object."""
        return self

    def __len__(self):
        """Returns the number of batches in the iterator."""
        return self.iterator_length

    def __del__(self):
        """Releases the rocAL resources."""
        for pipeline in self.pipelines:
            b.rocalRelease(pipeline._handle)


def get_spec_for_array(jax_array):
    """Returns the ArraySpec for a given JAX array.

    Args:
        jax_array (jax.Array): The JAX array.

    Returns:
        ArraySpec: The specification of the array.
    """
    return ArraySpec(shape=jax_array.shape, dtype=jax_array.dtype)


class ROCALPeekableIterator(ROCALJaxIterator):
    """ROCALJaxIterator extended with peek functionality. Compatible with Google CLU PeekableIterator.

     Reference: https://github.com/google/CommonLoopUtils/blob/main/clu/data/dataset_iterator.py
    """

    def __init__(self, pipelines, sharding=None):
        if not CLU_FOUND:
            print('Install CLU for peekable data iterator support')
            raise ImportError
        super().__init__(
            pipelines,
            sharding
        )
        self.mutex = threading.Lock()
        self.pool = None
        self.peek = None

        self.element_spec = None

    def set_element_spec(self, outputs):
        """Sets the element spec for the iterator.

        Args:
            outputs: The output from the iterator.
        """
        self.element_spec = [get_spec_for_array(output) for output in outputs]

    def assert_output_shape_and_type(self, outputs):
        """Asserts that the shape and type of the outputs are consistent.

        Args:
            outputs: The output from the iterator.

        Raises:
            ValueError: If the shape or type of the output changes between iterations.

        Returns:
            The outputs if they are consistent.
        """
        if self.element_spec is None:
            # Set element spec based on the first seen element
            self.set_element_spec(outputs)

        for idx, _ in enumerate(outputs):
            if get_spec_for_array(outputs[idx]) != self.element_spec[idx]:
                raise ValueError(
                    "The shape or type of the output changed between iterations. "
                    "This is not supported by JAX  peekable iterator. "
                    "Please make sure that the shape and type of the output is constant. "
                    f"Expected: {self.element_spec[idx]}, got: {get_spec_for_array(outputs[idx])} "
                    f"for output: {idx}"
                )

        return outputs

    def next_with_peek_impl(self):
        """Returns the next element from the iterator and advances the iterator.
        """
        if self.peek is None:
            return self.assert_output_shape_and_type(super().__next__())
        peek = self.peek
        self.peek = None
        return self.assert_output_shape_and_type(peek)

    def __next__(self):
        """Returns the next element from the iterator and advances it."""
        with self.mutex:
            return self.next_with_peek_impl()

    def __iter__(self):
        """Returns the iterator object. Resets if the iterator has been used."""
        if self._has_started and self.peek is None:
            self.reset()
        return self

    def peek(self):
        """Returns the next element from the iterator without advancing the iterator.
        """
        with self.mutex:
            if self.peek is None:
                self.peek = self.next_with_peek_impl()
            return self.peek

    def peek_async(self):
        """Returns future that will return the next element from
        the iterator without advancing the iterator.
        """
        if self.pool is None:
            # Create pool only if needed (peek_async is ever called)
            # to avoid thread creation overhead
            self.pool = concurrent.futures.ThreadPoolExecutor(max_workers=1)

        future = self.pool.submit(self.peek)
        return future

    @property
    def element_spec(self):
        """Returns the element spec for the elements returned by the iterator.
        """
        if self.element_spec is None:
            self.set_element_spec(self.peek())
        return self.element_spec



# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##
# @file decoders.py
#
# @brief  File containing various decoder implementations for various readers

import amd.rocal.types as types
import rocal_pybind as b
from amd.rocal.pipeline import Pipeline


def image(*inputs, user_feature_key_map=None, path='', file_root='', annotations_file='', index_path ='', shard_id=0, num_shards=1, random_shuffle=False,
          output_type=types.RGB, decoder_type=types.DECODER_TJPEG, device=None,
          decode_size_policy=types.USER_GIVEN_SIZE_ORIG, max_decoded_width=1000, max_decoded_height=1000,
          last_batch_policy=types.LAST_BATCH_FILL, pad_last_batch=True, stick_to_shard=True, shard_size=-1):
    """!Decodes images using different readers and decoders.

        @param inputs                   list of input images.
        @param user_feature_key_map     User-provided feature key mapping.
        @param path                     Path to image source.
        @param file_root                Root path for image files.
        @param annotations_file         Path to annotations file.
        @param shard_id                 Shard ID for parallel processing.
        @param num_shards               Total number of shards for parallel processing.
        @param random_shuffle           Whether to shuffle images randomly.
        @param output_type              Color format of the output image.
        @param decoder_type             Type of image decoder to use - TurboJpeg, OpenCV or rocJpeg
        @param device                   Device to use for decoding ("gpu" or "cpu").
        @param decode_size_policy       Size policy for decoding images.
        @param max_decoded_width        Maximum width for decoded images.
        @param max_decoded_height       Maximum height for decoded images.

        @return    Decoded and preprocessed image.
    """
    reader = Pipeline._current_pipeline._reader
    Pipeline._current_pipeline._last_batch_policy = last_batch_policy
    sharding_info = b.RocalShardingInfo(last_batch_policy, pad_last_batch, stick_to_shard, shard_size)

    if (device == "gpu"):
        decoder_type = types.DECODER_ROCJPEG
    if (reader == 'COCOReader'):
        kwargs_pybind = {
            "source_path": file_root,
            "json_path": annotations_file,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        decoded_image = b.cocoImageDecoderShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))

    elif (reader == "TFRecordReaderClassification" or reader == "TFRecordReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "num_shards": num_shards,
            'is_output': False,
            "user_key_for_encoded": user_feature_key_map["image/encoded"],
            "user_key_for_filename": user_feature_key_map["image/filename"],
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        decoded_image = b.tfImageDecoder(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))

    elif (reader == "Caffe2Reader" or reader == "Caffe2ReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        decoded_image = b.caffe2ImageDecoderShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))

    elif reader == "CaffeReader" or reader == "CaffeReaderDetection":
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        decoded_image = b.caffeImageDecoderShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))

    elif reader == "MXNETReader":
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        decoded_image = b.mxnetDecoder(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    elif reader == "WebDataset":
        kwargs_pybind = {
            "source_path": file_root,
            "index_path": index_path,
            "output_type": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        decoded_image = b.webdatasetSourceSingleShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    else:
        kwargs_pybind = {
            "source_path": file_root,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        decoded_image = b.imageDecoderShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))

    return (decoded_image)


def image_raw(*inputs, user_feature_key_map=None, path='', random_shuffle=False, output_type=types.RGB, max_decoded_width=1000, max_decoded_height=1000,
              last_batch_policy=types.LAST_BATCH_FILL, pad_last_batch=True, stick_to_shard=True, shard_size=-1):
    """!Decodes raw images for TF reader and decoder.

        @param inputs                  list of input images.
        @param user_feature_key_map    User-provided feature key mapping.
        @param path                    Path to image source.
        @param random_shuffle          Whether to shuffle images randomly.
        @param output_type             Color format of the output image.
        @param max_decoded_width       Maximum width for decoded images.
        @param max_decoded_height      Maximum height for decoded images.

        @return    Decoded raw image.
    """
    reader = Pipeline._current_pipeline._reader
    Pipeline._current_pipeline._last_batch_policy = last_batch_policy
    sharding_info = b.RocalShardingInfo(last_batch_policy, pad_last_batch, stick_to_shard, shard_size)

    if (reader == "TFRecordReaderClassification" or reader == "TFRecordReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "user_key_for_encoded": user_feature_key_map["image/encoded"],
            "user_key_for_filename": user_feature_key_map["image/filename"],
            "color_format": output_type,
            'is_output': False,
            "shuffle": random_shuffle,
            "loop": False,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info}
        decoded_image = b.tfImageDecoderRaw(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
        return (decoded_image)


def image_random_crop(*inputs, user_feature_key_map=None, path='', file_root='', annotations_file='', num_shards=1, shard_id=0,
                      random_shuffle=False, num_attempts=10, output_type=types.RGB, random_area=[0.08, 1.0],
                      random_aspect_ratio=[0.8, 1.25], decode_size_policy=types.USER_GIVEN_SIZE_ORIG,
                      max_decoded_width=1000, max_decoded_height=1000, decoder_type=types.DECODER_TJPEG,
                      last_batch_policy=types.LAST_BATCH_FILL, pad_last_batch=True, stick_to_shard=True, shard_size=-1):
    """!Applies random cropping to images using different readers and decoders.

        @param inputs                  list of input images.
        @param user_feature_key_map    User-provided feature key mapping.
        @param path                    Path to image source.
        @param file_root               Root path for image files.
        @param annotations_file        Path to annotations file.
        @param num_shards              Total number of shards for parallel processing.
        @param shard_id                Shard ID for parallel processing.
        @param random_shuffle          Whether to shuffle images randomly.
        @param num_attempts            Maximum number of attempts to find a valid crop.
        @param output_type             Color format of the output image.
        @param random_area             Random areas for cropping.
        @param random_aspect_ratio     Random aspect ratios for cropping.
        @param decode_size_policy      Size policy for decoding images.
        @param max_decoded_width       Maximum width for decoded images.
        @param max_decoded_height      Maximum height for decoded images.
        @param decoder_type            Type of image decoder to use.

        @return    Randomly cropped and preprocessed image.
    """
    reader = Pipeline._current_pipeline._reader
    Pipeline._current_pipeline._last_batch_policy = last_batch_policy
    sharding_info = b.RocalShardingInfo(last_batch_policy, pad_last_batch, stick_to_shard, shard_size)

    # Internally calls the C++ Partial decoder's
    if (reader == 'COCOReader'):
        kwargs_pybind = {
            "source_path": file_root,
            "json_path": annotations_file,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info}
        crop_output_image = b.cocoImageDecoderSliceShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    elif (reader == "TFRecordReaderClassification" or reader == "TFRecordReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "num_shards": num_shards,
            'is_output': False,
            "user_key_for_encoded": user_feature_key_map["image/encoded"],
            "user_key_for_filename": user_feature_key_map["image/filename"],
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "dec_type": decoder_type,
            "sharding_info": sharding_info}
        crop_output_image = b.tfImageDecoder(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    elif (reader == "CaffeReader" or reader == "CaffeReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info}
        crop_output_image = b.caffeImageDecoderPartialShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    elif (reader == "Caffe2Reader" or reader == "Caffe2ReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info}
        crop_output_image = b.caffe2ImageDecoderPartialShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    else:
        kwargs_pybind = {
            "source_path": file_root,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info}
        crop_output_image = b.fusedDecoderCropShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))

    return (crop_output_image)


def image_slice(*inputs, file_root='', path='', annotations_file='', shard_id=0, num_shards=1, random_shuffle=False, device = 'cpu',
                random_aspect_ratio=[0.75, 1.33333], random_area=[0.08, 1.0], num_attempts=100, output_type=types.RGB,
                decode_size_policy=types.USER_GIVEN_SIZE_ORIG, max_decoded_width=1000, max_decoded_height=1000, last_batch_policy=types.LAST_BATCH_FILL, pad_last_batch=True, stick_to_shard=True, shard_size=-1):
    """!Slices images randomly using different readers and decoders.

        @param inputs                 list of input images.
        @param file_root              Root path for image files.
        @param path                   Path to image source.
        @param annotations_file       Path to annotations file.
        @param shard_id               Shard ID for parallel processing.
        @param num_shards             Total number of shards for parallel processing.
        @param random_shuffle         Whether to shuffle images randomly.
        @param random_aspect_ratio    Random aspect ratios for cropping.
        @param random_area            Random areas for cropping.
        @param num_attempts           Maximum number of attempts to find a valid crop.
        @param output_type            Color format of the output image.
        @param decode_size_policy     Size policy for decoding images.
        @param max_decoded_width      Maximum width for decoded images.
        @param max_decoded_height     Maximum height for decoded images.

        @return    Sliced image.
    """
    reader = Pipeline._current_pipeline._reader
    Pipeline._current_pipeline._last_batch_policy = last_batch_policy
    sharding_info = b.RocalShardingInfo(last_batch_policy, pad_last_batch, stick_to_shard, shard_size)
    
    if (device == "gpu"):
        decoder_type = types.DECODER_ROCJPEG
    else:
        decoder_type = types.DECODER_TJPEG
    # Reader -> Randon BBox Crop -> ImageDecoderSlice
    # Random crop parameters taken from pytorch's RandomResizedCrop default function arguments
    # TODO:To pass the crop co-ordinates from random_bbox_crop to image_slice
    # in tensor branch integration,
    # for now calling partial decoder to match SSD training outer API's .
    if (reader == 'COCOReader'):
        kwargs_pybind = {
            "source_path": file_root,
            "json_path": annotations_file,
            "color_format": output_type,
            "shard_id": shard_id,
            "shard_count": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info,
            "dec_type": decoder_type}
        image_decoder_slice = b.cocoImageDecoderSliceShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    elif (reader == "CaffeReader" or reader == "CaffeReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info,
            "dec_type": decoder_type}
        image_decoder_slice = b.caffeImageDecoderPartialShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    elif (reader == "Caffe2Reader" or reader == "Caffe2ReaderDetection"):
        kwargs_pybind = {
            "source_path": path,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info,
            "dec_type": decoder_type}
        image_decoder_slice = b.caffe2ImageDecoderPartialShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    else:
        kwargs_pybind = {
            "source_path": file_root,
            "color_format": output_type,
            "shard_id": shard_id,
            "num_shards": num_shards,
            'is_output': False,
            "area_factor": random_area,
            "aspect_ratio": random_aspect_ratio,
            "num_attempts": num_attempts,
            "shuffle": random_shuffle,
            "loop": False,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_width,
            "max_height": max_decoded_height,
            "sharding_info": sharding_info,
            "dec_type": decoder_type}
        image_decoder_slice = b.fusedDecoderCropShard(
            Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    return (image_decoder_slice)

def audio(*inputs, file_root='', file_list_path='', bytes_per_sample_hint=[0], shard_id=0, num_shards=1, random_shuffle=False, downmix=False, dtype=types.FLOAT, quality=50.0, sample_rate=0.0, seed=1, stick_to_shard=True, shard_size=-1, last_batch_policy=types.LAST_BATCH_FILL, pad_last_batch_repeated=False,
          decode_size_policy=types.MAX_SIZE, max_decoded_samples=522320, max_decoded_channels=1):
    """!Decodes wav audio files.

        @param inputs                   list of input audio.
        @param file_root                Folder Path to the audio data.
        @param file_list_path           Text file containing list of files and the labels
        @param shard_id                 Shard ID for parallel processing.
        @param num_shards               Total number of shards for parallel processing.
        @param random_shuffle           Whether to shuffle images randomly.
        @param downmix                  Converts the audio data to single channel when enabled 
        @param dtype                    Data type of the decoded audio
        @param quality                  Resampling quality
        @param sample_rate              Sample rate for the decoded audio.
        @param seed                     Random seed.
        @param stick_to_shard           The reader sticks to the data for it's corresponding shard when enabled
        @param shard_size               Number of files in an epoch
        @param last_batch_policy        Determines the handling of the last batch when the shard size is not divisible by the batch size. Check types.py enum for possible values.
        @param pad_last_batch_repeated  If set to True, pads the shards last batch by repeating the last sample's data (dummy data).
        @param decode_size_policy       Size policy for decoding images.
        @param max_decoded_samples      Maximum samples for decoded images.
        @param max_decoded_channels     Maximum channels for decoded images.
        @return                         Decoded audio.
    """
    sharding_info = b.RocalShardingInfo(last_batch_policy, pad_last_batch_repeated, stick_to_shard, shard_size)
    kwargs_pybind = {
            "source_path": file_root,
            "source_file_list_path": file_list_path,
            "shard_id": shard_id,
            "num_shards": num_shards,
            "is_output": False,
            "shuffle": random_shuffle,
            "loop": False,
            "downmix": downmix,
            "decode_size_policy": decode_size_policy,
            "max_width": max_decoded_samples,
            "max_height": max_decoded_channels,
            "sharding_info": sharding_info}
    Pipeline._current_pipeline._last_batch_policy = last_batch_policy
    decoded_audio = b.audioDecoderSingleShard(Pipeline._current_pipeline._handle, *(kwargs_pybind.values()))
    return decoded_audio




# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##
# @file pytorch.py
# @brief File containing iterators to be used with pytorch trainings

import torch
import numpy as np
import rocal_pybind as b
import amd.rocal.types as types
import ctypes


class ROCALGenericIterator(object):
    """!Iterator for processing data

        @param pipeline            The rocAL pipeline to use for processing data.
        @param tensor_layout       The layout of the output tensors
        @param reverse_channels    Whether to reverse the order of color channels.
        @param multiplier          Multiplier values for color normalization.
        @param offset              Offset values for color normalization.
        @param tensor_dtype        Data type of the output tensors.
        @param display             Whether to display images during processing
        @param device              The device to use for processing
        @param device_id           The ID of the device to use
    """

    def __init__(self, pipeline, tensor_layout=types.NCHW, reverse_channels=False, multiplier=[1.0, 1.0, 1.0], offset=[0.0, 0.0, 0.0], tensor_dtype=types.FLOAT, device="cpu", device_id=0, display=False):
        self.loader = pipeline
        self.tensor_format = tensor_layout
        self.multiplier = multiplier
        self.offset = offset
        self.reverse_channels = reverse_channels
        self.tensor_dtype = tensor_dtype
        self.device = device
        self.device_id = device_id
        self.batch_size = self.loader._batch_size
        self.labels_size = ((self.batch_size * self.loader._num_classes)
                            if self.loader._one_hot_encoding else self.batch_size)
        self.output_list = None
        self.output_memory_type = self.loader._output_memory_type
        self.iterator_length = b.getRemainingImages(self.loader._handle)
        self.display = display
        self.batch_size = pipeline._batch_size
        if self.loader._is_external_source_operator:
            self.eos = False
            self.index = 0
            self.num_batches = self.loader._external_source.n // self.batch_size if self.loader._external_source.n % self.batch_size == 0 else (
                self.loader._external_source.n // self.batch_size + 1)
        else:
            self.num_batches = None
        if self.loader._name is None:
            self.loader._name = self.loader._reader
        self.last_batch_policy = self.loader._last_batch_policy
        self.last_batch_size = None

    def next(self):
        return self.__next__()

    def __next__(self):
        if (self.loader._is_external_source_operator):
            if (self.index + 1) == self.num_batches:
                self.eos = True
            if (self.index + 1) <= self.num_batches:
                data_loader_source = next(self.loader._external_source)
                # Extract all data from the source
                images_list = data_loader_source[0] if (self.loader._external_source_mode == types.EXTSOURCE_FNAME) else []
                input_buffer = data_loader_source[0] if (self.loader._external_source_mode != types.EXTSOURCE_FNAME) else []
                labels_data = data_loader_source[1] if (len(data_loader_source) > 1) else None
                roi_height = data_loader_source[2] if (len(data_loader_source) > 2) else []
                roi_width = data_loader_source[3] if (len(data_loader_source) > 3) else []
                if (len(data_loader_source) == 6 and self.loader._external_source_mode == types.EXTSOURCE_RAW_UNCOMPRESSED):
                    decoded_height = data_loader_source[4]
                    decoded_width = data_loader_source[5]
                else:
                    decoded_height = self.loader._external_source_user_given_height
                    decoded_width = self.loader._external_source_user_given_width

                kwargs_pybind = {
                    "handle": self.loader._handle,
                    "source_input_images": images_list,
                    "labels": labels_data,
                    "input_batch_buffer": input_buffer,
                    "roi_width": roi_width,
                    "roi_height": roi_height,
                    "decoded_width": decoded_width,
                    "decoded_height": decoded_height,
                    "channels": 3,
                    "external_source_mode": self.loader._external_source_mode,
                    "rocal_tensor_layout": types.NCHW,
                    "eos": self.eos}
                b.externalSourceFeedInput(*(kwargs_pybind.values()))
            self.index = self.index + 1
        if self.loader.rocal_run() != 0:
            b.rocalResetLoaders(self.loader._handle)
            raise StopIteration
        else:
            self.output_tensor_list = self.loader.get_output_tensors()

        if self.output_list is None:
            # Output list used to store pipeline outputs - can support multiple augmentation outputs
            self.output_list = []
            for i in range(len(self.output_tensor_list)):
                dimensions = self.output_tensor_list[i].dimensions()
                if self.device == "cpu":
                    torch_dtype = self.output_tensor_list[i].dtype()
                    output = torch.empty(
                        dimensions, dtype=getattr(torch, torch_dtype))
                    self.labels_tensor = torch.empty(
                        self.labels_size, dtype=getattr(torch, "int32"))
                else:
                    torch_gpu_device = torch.device('cuda', self.device_id)
                    torch_dtype = self.output_tensor_list[i].dtype()
                    output = torch.empty(dimensions, dtype=getattr(
                        torch, torch_dtype), device=torch_gpu_device)
                    self.labels_tensor = torch.empty(self.labels_size, dtype=getattr(
                        torch, "int32"), device=torch_gpu_device)

                self.output_tensor_list[i].copy_data(ctypes.c_void_p(
                    output.data_ptr()), self.output_memory_type)
                self.output_list.append(output)
        else:
            for i in range(len(self.output_tensor_list)):
                self.output_tensor_list[i].copy_data(ctypes.c_void_p(
                    self.output_list[i].data_ptr()), self.output_memory_type)

        if ((self.loader._name == "Caffe2ReaderDetection") or (self.loader._name == "CaffeReaderDetection")):
            self.bbox_list = []  # Empty list for bboxes
            self.labels_list = []  # Empty list of labels

            # 1D labels array in a batch
            self.labels = self.loader.get_bounding_box_labels()
            # 1D bboxes array in a batch
            self.bboxes = self.loader.get_bounding_box_cords()
            # Image sizes of a batch
            self.img_size = np.zeros((self.batch_size * 2), dtype="int32")
            self.loader.get_img_sizes(self.img_size)

            for i in range(self.batch_size):
                self.label_2d_numpy = np.reshape(
                    self.labels[i], (-1, 1)).tolist()
                self.bb_2d_numpy = np.reshape(self.bboxes[i], (-1, 4)).tolist()

                self.labels_list.append(self.label_2d_numpy)
                self.bbox_list.append(self.bb_2d_numpy)

                if self.display:
                    for output in self.output_list:
                        img = output
                        draw_patches(img[i], i, self.bb_2d_numpy)

            max_cols = max([len(row)
                           for batch in self.bbox_list for row in batch])
            max_rows = max([len(batch) for batch in self.bbox_list])
            self.bb_padded = [
                batch + [[0] * (max_cols)] * (max_rows - len(batch)) for batch in self.bbox_list]
            self.bb_padded = torch.FloatTensor(
                [row + [0] * (max_cols - len(row)) for batch in self.bb_padded for row in batch])
            self.bb_padded = self.bb_padded.view(-1, max_rows, max_cols)

            max_cols1 = max([len(row)
                            for batch in self.labels_list for row in batch])
            max_rows1 = max([len(batch) for batch in self.labels_list])
            self.labels_padded = [
                batch + [[0] * (max_cols1)] * (max_rows1 - len(batch)) for batch in self.labels_list]
            self.labels_padded = torch.LongTensor(
                [row + [0] * (max_cols1 - len(row)) for batch in self.labels_padded for row in batch])
            self.labels_padded = self.labels_padded.view(
                -1, max_rows1, max_cols1)

            # Check if last batch policy is partial and only return the valid images in last batch
            if (self.last_batch_policy is (types.LAST_BATCH_PARTIAL)) and b.getRemainingImages(self.loader._handle) <= 0:
                if (self.last_batch_size is None):
                    self.last_batch_size = self.batch_size - \
                        b.getLastBatchPaddedSize(self.loader._handle)
                return [inner_list[0:self.last_batch_size, :] for inner_list in self.output_list], self.bb_padded[0:self.last_batch_size], self.labels_padded[0:self.last_batch_size]
            else:
                return self.output_list, self.bb_padded, self.labels_padded

        elif self.loader._is_external_source_operator:
            self.labels = self.loader.get_image_labels()
            self.labels_tensor = self.labels_tensor.copy_(
                torch.from_numpy(self.labels)).long()
            return self.output_list, self.labels_tensor
        else:
            if self.loader._one_hot_encoding:
                self.loader.get_one_hot_encoded_labels(
                    self.labels_tensor.data_ptr(), self.loader._output_memory_type)
                self.labels_tensor = self.labels_tensor.reshape(
                    -1, self.batch_size, self.loader._num_classes)
            else:
                if self.display:
                    for i in range(self.batch_size):
                        img = (self.output_list[0])
                        draw_patches(img[i], i, [])
                if (self.loader._name == "WebDataset"):
                    self.ascii_outputs = self.loader.get_ascii_datas()
                else:
                    self.labels = self.loader.get_image_labels()
                    self.labels_tensor = self.labels_tensor.copy_(
                        torch.from_numpy(self.labels)).long()

            # Check if last batch policy is partial and only return the valid images in last batch
            if (self.last_batch_policy is (types.LAST_BATCH_PARTIAL)) and b.getRemainingImages(self.loader._handle) < self.batch_size:
                if (self.last_batch_size is None):
                    self.last_batch_size = self.batch_size - \
                        b.getLastBatchPaddedSize(self.loader._handle)
                if (self.loader._name == "WebDataset"):
                    return [inner_list[0:self.last_batch_size, :] for inner_list in self.output_list], [inner_list[0:self.last_batch_size] for inner_list in self.ascii_outputs]
                else:
                    return [inner_list[0:self.last_batch_size, :] for inner_list in self.output_list], self.labels_tensor[0:self.last_batch_size]
            else:
                return self.output_list, self.ascii_outputs if (self.loader._name == "WebDataset") else self.labels_tensor

    def reset(self):
        b.rocalResetLoaders(self.loader._handle)

    def __iter__(self):
        return self

    def __len__(self):
        return self.iterator_length // self.batch_size

    def __del__(self):
        b.rocalRelease(self.loader._handle)


class ROCALClassificationIterator(ROCALGenericIterator):
    """!rocAL iterator for classification tasks for PyTorch. It returns 2 outputs
    (data and label) in the form of PyTorch's Tensors.

    Calling

    .. code-block:: python

       ROCALClassificationIterator(pipelines, size)

    is equivalent to calling

    .. code-block:: python

       ROCALGenericIterator(pipelines, ["data", "label"], size)

    Please keep in mind that Tensors returned by the iterator are
    still owned by rocAL. They are valid till the next iterator call.
    If the content needs to be preserved please copy it to another tensor.

    pipelines (list of amd.rocal.pipeline.Pipeline)       List of pipelines to use
    size (int)                                            Number of samples in the epoch (Usually the size of the dataset).
    auto_reset (bool, optional, default = False)          Whether the iterator resets itself for the next epoch or it requires reset() to be called separately.
    fill_last_batch (bool, optional, default = True)      Whether to fill the last batch with data up to 'self.batch_size'. The iterator would return the first integer multiple of self._num_gpus * self.batch_size entries which exceeds 'size'. Setting this flag to False will cause the iterator to return exactly 'size' entries.
    dynamic_shape (bool, optional, default = False)       Whether the shape of the output of the rocAL pipeline can change during execution. If True, the pytorch tensor will be resized accordingly if the shape of rocAL returned tensors changes during execution. If False, the iterator will fail in case of change.
    last_batch_padded (bool, optional, default = False)   Whether the last batch provided by rocAL is padded with the last sample or it just wraps up. In the conjunction with fill_last_batch it tells if the iterator returning last batch with data only partially filled with data from the current epoch is dropping padding samples or samples from the next epoch. If set to False next epoch will end sooner as data from it was consumed but dropped. If set to True next epoch would be the same length as the first one.

    Example
    -------
    With the data set [1,2,3,4,5,6,7] and the batch size 2:
    fill_last_batch = False, last_batch_padded = True  -> last batch = [7], next iteration will return [1, 2]
    fill_last_batch = False, last_batch_padded = False -> last batch = [7], next iteration will return [2, 3]
    fill_last_batch = True, last_batch_padded = True   -> last batch = [7, 7], next iteration will return [1, 2]
    fill_last_batch = True, last_batch_padded = False  -> last batch = [7, 1], next iteration will return [2, 3]
    """

    def __init__(self,
                 pipelines,
                 size=0,
                 auto_reset=False,
                 fill_last_batch=True,
                 dynamic_shape=False,
                 last_batch_padded=False,
                 display=False,
                 device="cpu",
                 device_id=0):
        pipe = pipelines
        super(ROCALClassificationIterator, self).__init__(pipe, tensor_layout=pipe._tensor_layout, tensor_dtype=pipe._tensor_dtype,
                                                          multiplier=pipe._multiplier, offset=pipe._offset, display=display, device=device, device_id=device_id)


class ROCALAudioIterator(object):
    """! rocAL iterator for audio tasks for PyTorch
    The Tensors that are returned by the iterator will be owned by rocAL and would be valid until next iteration.
        @param pipeline            The rocAL pipeline to use for processing data.
        @param tensor_dtype        Data type of the output tensors.
        @param size                Number of samples in the epoch (Usually the size of the dataset).
        @param auto_reset          Whether the iterator resets itself for the next epoch or it requires reset() to be called separately.
        @param device              The device to use for processing - CPU / GPU
        @param device_id           The ID of the device to use
    """

    def __init__(self, pipeline, tensor_dtype=types.FLOAT, size=-1, auto_reset=False, device="cpu", device_id=0):
        self.loader = pipeline
        self.device = device
        self.device_id = device_id
        self.output = None
        # To change the name of getRemainingImages to getRemainingSamples in upcoming PRs
        self.iterator_length = b.getRemainingImages(self.loader._handle)
        self.max_shape = None
        self.batch_size = self.loader._batch_size
        self.output_list = None
        self.labels_size = self.batch_size
        self.output_memory_type = self.loader._output_memory_type
        self.last_batch_padded_size = b.getLastBatchPaddedSize(
            self.loader._handle)
        self.last_batch_policy = self.loader._last_batch_policy
        self.shard_size = size
        self.auto_reset = auto_reset
        self.batch_count = 0

    def next(self):
        return self.__next__()

    def __next__(self):
        rocal_run = self.loader.rocal_run()
        if (rocal_run != 0 or self.shard_size > 0 and self.batch_count >= self.shard_size):
            if self.auto_reset:
                self.reset()
            raise StopIteration
        else:
            self.output_tensor_list = self.loader.get_output_tensors()
        self.batch_count += self.batch_size
        self.last_batch_size = self.batch_size - \
            b.getLastBatchPaddedSize(self.loader._handle)
        self.output_list = []
        for i in range(len(self.output_tensor_list)):
            dimensions = self.output_tensor_list[i].dimensions()
            self.num_roi_dims = self.output_tensor_list[i].roi_dims_size()
            self.num_of_rois = self.num_roi_dims * 2
            self.roi_array = np.zeros(
                self.batch_size * self.num_of_rois, dtype=np.int32)
            self.output_tensor_list[i].copy_roi(self.roi_array)
            roi = self.roi_array.reshape(self.batch_size, self.num_of_rois)
            max_x1 = np.max(roi[..., self.num_of_rois-2:self.num_of_rois-1])
            max_y1 = np.max(roi[..., self.num_of_rois-1:self.num_of_rois])
            if self.device == "cpu":
                torch_dtype = self.output_tensor_list[i].dtype()
                if (max_x1 == 0 or max_y1 == 0):
                    output = torch.empty(
                        dimensions, dtype=getattr(torch, torch_dtype))
                else:
                    output = torch.empty(
                        dimensions[0], max_x1, max_y1, dtype=getattr(torch, torch_dtype))
                self.labels_tensor = torch.empty(
                    self.labels_size, dtype=getattr(torch, torch_dtype))
            else:
                torch_gpu_device = torch.device('cuda', self.device_id)
                torch_dtype = self.output_tensor_list[i].dtype()
                if (max_x1 == 0 or max_y1 == 0):
                    output = torch.empty(dimensions, dtype=getattr(
                        torch, torch_dtype), device=torch_gpu_device)
                else:
                    output = torch.empty(
                        dimensions[0], max_x1, max_y1, dtype=getattr(torch, torch_dtype))
                self.labels_tensor = torch.empty(self.labels_size, dtype=getattr(
                    torch, torch_dtype), device=torch_gpu_device)

            if (max_x1 == 0 or max_y1 == 0):
                self.output_tensor_list[i].copy_data(ctypes.c_void_p(
                    output.data_ptr()), self.output_memory_type)
            else:
                self.output_tensor_list[i].copy_data(
                    ctypes.c_void_p(output.data_ptr()), 0, 0, max_y1, max_x1)
            self.output_list.append(output)

        self.labels = self.loader.get_image_labels()
        self.labels_tensor = self.labels_tensor.copy_(
            torch.from_numpy(self.labels)).long()

        if (self.last_batch_policy is (types.LAST_BATCH_PARTIAL)) and b.getRemainingImages(self.loader._handle) < self.batch_size:
            return [inner_list[0:self.last_batch_size, :] for inner_list in self.output_list], self.labels_tensor[0:self.last_batch_size], torch.from_numpy(self.roi_array.reshape(self.batch_size, self.num_of_rois)[..., self.num_of_rois-2:self.num_of_rois][0:self.last_batch_size, :])
        else:
            return self.output_list, self.labels_tensor, torch.from_numpy(self.roi_array.reshape(self.batch_size, self.num_of_rois)[..., self.num_of_rois-2:self.num_of_rois])

    def reset(self):
        self.batch_count = 0
        b.rocalResetLoaders(self.loader._handle)

    def __iter__(self):
        return self

    def __len__(self):
        return self.iterator_length

    def __del__(self):
        b.rocalRelease(self.loader._handle)


class ROCALNumpyIterator(object):
    def __init__(self, pipeline, device="cpu", device_id=0):
        self.loader = pipeline
        self.device = device
        self.device_id = device_id
        self.output_memory_type = self.loader._output_memory_type
        self.output_list = None
        self.batch_size = self.loader._batch_size
        self.len = b.getRemainingImages(self.loader._handle) // self.batch_size
        self.last_batch_policy = self.loader._last_batch_policy
        self.last_batch_size = None

    def next(self):
        return self.__next__()

    def __next__(self):
        if self.loader.rocal_run() != 0:
            raise StopIteration
        self.output_tensor_list = self.loader.get_output_tensors()

        if self.output_list is None:
            # Output list used to store pipeline outputs - can support multiple augmentation outputs
            self.output_list = []
            for i in range(len(self.output_tensor_list)):
                dimensions = self.output_tensor_list[i].dimensions()
                torch_dtype = self.output_tensor_list[i].dtype()
                if self.device == "cpu":
                    output = torch.empty(
                        dimensions, dtype=getattr(torch, torch_dtype))
                else:
                    torch_gpu_device = torch.device('cuda', self.device_id)
                    output = torch.empty(dimensions, dtype=getattr(
                        torch, torch_dtype), device=torch_gpu_device)

                self.output_tensor_list[i].copy_data(ctypes.c_void_p(
                    output.data_ptr()), self.output_memory_type)
                self.output_list.append(output)
        else:
            for i in range(len(self.output_tensor_list)):
                self.output_tensor_list[i].copy_data(ctypes.c_void_p(
                    self.output_list[i].data_ptr()), self.output_memory_type)
        # Check if last batch policy is partial and only return the valid images in last batch
        if (self.last_batch_policy is (types.LAST_BATCH_PARTIAL)) and b.getRemainingImages(self.loader._handle) < self.batch_size:
            if (self.last_batch_size is None):
                self.last_batch_size = self.batch_size - \
                    b.getLastBatchPaddedSize(self.loader._handle)
            return [inner_list[0:self.last_batch_size, :] for inner_list in self.output_list]
        return self.output_list

    def reset(self):
        b.rocalResetLoaders(self.loader._handle)

    def __iter__(self):
        return self

    def __len__(self):
        return self.len

    def __del__(self):
        b.rocalRelease(self.loader._handle)

def draw_patches(img, idx, bboxes):
    """!Writes images to disk as a PNG file.

        @param img       The input image as a tensor.
        @param idx       Index used for naming the output file.
        @param bboxes    List of bounding boxes.
    """
    # image is expected as a tensor
    import cv2
    img = img.cpu()
    image = img.detach().numpy()
    image = image.transpose([1, 2, 0])
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    image = cv2.UMat(image).get()
    cv2.imwrite(str(idx)+"_"+"train"+".png", image)
    try:
        path = "OUTPUT_IMAGES_PYTHON/PYTORCH/"
        isExist = os.path.exists(path)
        if not isExist:
            os.makedirs(path)
    except OSError as error:
        print(error)
    if bboxes:
        for (l, t, r, b) in bboxes:
            loc_ = [l, t, r, b]
            color = (255, 0, 0)
            thickness = 2
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
            image = cv2.UMat(image).get()
            image = cv2.rectangle(image, (int(loc_[0]), int(loc_[1])), (int(
                (loc_[2])), int((loc_[3]))), color, thickness)
            cv2.imwrite("OUTPUT_IMAGES_PYTHON/PYTORCH/" +
                        str(idx) + "_" + "train" + ".png", image * 255)
    else:
        cv2.imwrite("OUTPUT_IMAGES_PYTHON/PYTORCH/" + str(idx) +
                    "_" + "train" + ".png", image * 255)



# Copyright (c) 2024 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from amd.rocal.pipeline import Pipeline
from amd.rocal.plugin.generic import ROCALNumpyIterator
import amd.rocal.fn as fn
import amd.rocal.types as types

import os
import shutil
import cv2
import numpy as np
from parse_config import parse_args


def draw_patches(image, idx, args=None):
    # image is expected as a numpy array
    if not args.NHWC:
        image = image.transpose([1, 2, 0])
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    cv2.imwrite("output_folder/numpy_reader/" + str(idx) + ".png", image)


def main():
    args = parse_args()
    # Args
    data_path = args.image_dataset_path
    rocal_cpu = False if args.rocal_gpu else True
    batch_size = args.batch_size
    num_threads = args.num_threads
    random_seed = args.seed
    local_rank = args.local_rank
    world_size = args.world_size
    output_layout = types.NHWC if args.NHWC else types.NCHW
    augmentation_name = args.augmentation_name

    try:
        path = "output_folder/numpy_reader/"
        if not os.path.exists(path):
            os.makedirs(path)
    except OSError as error:
        print(error)

    if augmentation_name == "log1p":
        # Create int16 numpy files since log1p only supports int16 inputs
        new_data_path = 'log1p_inputs'
        if os.path.exists(new_data_path):
            shutil.rmtree(new_data_path)
        os.makedirs(new_data_path)
        for i in range(10):
            np_array = np.random.randn(128, 128, 128, 4)
            np.save(os.path.join(new_data_path, f'{i}.npy'), np_array.astype(np.int16))
        data_path = new_data_path

    pipeline = Pipeline(batch_size=batch_size, num_threads=num_threads,
                        device_id=local_rank, seed=random_seed, rocal_cpu=rocal_cpu)

    with pipeline:
        numpy_reader_output = fn.readers.numpy(
            file_root=data_path, shard_id=local_rank, num_shards=world_size, output_layout=types.NHWC)
        if augmentation_name == "log1p":
            output = fn.log1p(numpy_reader_output)
        else:
            output = fn.resize(
                numpy_reader_output, resize_width=400, resize_height=400, output_layout=output_layout)
        pipeline.set_outputs(output)

    pipeline.build()

    cnt = 0
    numpyIteratorPipeline = ROCALNumpyIterator(pipeline)
    for epoch in range(args.num_epochs):
        print("epoch:: ", epoch)
        for i, [batch] in enumerate(numpyIteratorPipeline):
            if i == 0 and args.print_tensor:
                print(batch)
            for img in batch:
                if args.display:
                    draw_patches(img, cnt, args)
                cnt += 1
        numpyIteratorPipeline.reset()
    print("##############################  NUMPY READER SUCCESS  ############################")


if __name__ == '__main__':
    main()



# Copyright (c) 2018 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from amd.rocal.plugin.tf import ROCALIterator
from amd.rocal.pipeline import Pipeline
import amd.rocal.types as types
import os
import amd.rocal.fn as fn
import tensorflow as tf
import numpy as np
from parse_config import parse_args


def draw_patches(img, idx, device, args=None):
    import cv2
    args = parse_args()
    # converting to numpy for opencv display since iterator returns tf.tensor
    img = img.numpy()
    if not args.NHWC:
        img = img.transpose([0, 1, 2])
    image = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    cv2.imwrite("output_folder/tf_reader/classification/" +
                str(idx) + "_" + "train" + ".png", image)


def main():
    args = parse_args()
    # Args
    image_path = args.image_dataset_path
    rocal_cpu = False if args.rocal_gpu else True
    device = "cpu" if rocal_cpu else "gpu"
    batch_size = args.batch_size
    one_hot_labels = 1 if args.one_hot_encode else 0
    num_threads = args.num_threads
    tensor_layout = types.NHWC if args.NHWC else types.NCHW
    tf_record_reader_type = 0
    feature_key_map = {
        'image/encoded': 'image/encoded',
        'image/class/label': 'image/class/label',
        'image/filename': 'image/filename'
    }
    try:
        path = "output_folder/tf_reader/classification/"
        is_exist = os.path.exists(path)
        if not is_exist:
            os.makedirs(path)
    except OSError as error:
        print(error)
    # Create Pipeline instance
    pipe = Pipeline(batch_size=batch_size, num_threads=num_threads,
                    device_id=args.local_rank, seed=2, rocal_cpu=rocal_cpu)
    # Use pipeline instance to make calls to reader, decoder & augmentation's
    with pipe:
        inputs = fn.readers.tfrecord(path=image_path, reader_type=tf_record_reader_type, user_feature_key_map=feature_key_map,
                                     features={
                                         "image/encoded": tf.io.FixedLenFeature((), tf.string, ""),
                                         "image/class/label": tf.io.FixedLenFeature([1], tf.int64, -1),
                                         "image/filename": tf.io.FixedLenFeature((), tf.string, "")
                                     }
                                     )
        jpegs = inputs["image/encoded"]
        images = fn.decoders.image(
            jpegs, user_feature_key_map=feature_key_map, output_type=types.RGB, path=image_path)
        resized = fn.resize(images, resize_width=300,
                            resize_height=300, output_layout=tensor_layout)
        if one_hot_labels == 1:
            labels = inputs["image/class/label"]
            _ = fn.one_hot(labels, num_classes=1000)
        pipe.set_outputs(resized)
    # Build the pipeline
    pipe.build()
    # Dataloader
    image_iterator = ROCALIterator(pipe, device=device)
    cnt = 0
    # Enumerate over the Dataloader
    for i, ([images_array], labels_array) in enumerate(image_iterator, 0):
        if args.print_tensor:
            print("\n", i)
            print("lables_array", labels_array)
            print("\n\nPrinted first batch with", (batch_size), "images!")
        for element in list(range(batch_size)):
            cnt += 1
            draw_patches(images_array[element], cnt, device, args=args)
        break
    image_iterator.reset()

    print("##############################  TF CLASSIFICATION  SUCCESS  ############################")


if __name__ == "__main__":
    main()



from random import shuffle
from amd.rocal.pipeline import Pipeline
from amd.rocal.plugin.generic import ROCALClassificationIterator
import amd.rocal.fn as fn
import amd.rocal.types as types
import os
import numpy as np
import cv2
import sys
import glob


def main():
    if  len(sys.argv) < 3:
        print ('Please pass cpu/gpu batch_size')
        exit(0)
    batch_size = int(sys.argv[2])
    device = "cpu" if sys.argv[1] == "cpu" else "gpu"
    data_dir = os.environ["ROCAL_DATA_PATH"] + \
        "rocal_data/coco/coco_10_img/images/"
    try:
        path_mode0 = "output_folder/external_source_reader/mode0/"
        isExist = os.path.exists(path_mode0)
        if not isExist:
            os.makedirs(path_mode0)
    except OSError as error:
        print(error)
    try:
        path_mode1 = "output_folder/external_source_reader/mode1/"
        isExist = os.path.exists(path_mode1)
        if not isExist:
            os.makedirs(path_mode1)
    except OSError as error:
        print(error)
    try:
        path_mode2 = "output_folder/external_source_reader/mode2/"
        isExist = os.path.exists(path_mode2)
        if not isExist:
            os.makedirs(path_mode2)
    except OSError as error:
        print(error)

    def image_dump(img, idx, mode=0):
        img = img.transpose([1, 2, 0])  # NCHW
        img = (img).astype('uint8')
        if mode!=2:
            img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        cv2.imwrite("output_folder/external_source_reader/mode" + str(mode) + "/"+
                    str(idx)+"_"+"train"+".png", img)

    ##################### MODE 0 #########################
    # Define the Data Source for all image samples - User needs to define their own source
    class ExternalInputIteratorMode0(object):
        def __init__(self, batch_size):
            self.images_dir = data_dir
            self.batch_size = batch_size
            self.files = []
            self.file_patterns = ['*.jpg', '*.jpeg', '*.JPG', '*.JPEG']
            for pattern in self.file_patterns:
                self.files.extend(glob.glob(os.path.join(self.images_dir, pattern)))
            shuffle(self.files)

        def __iter__(self):
            self.i = 0
            self.n = len(self.files)
            return self

        def __next__(self):
            batch = []
            labels = []
            label = 1
            for _ in range(self.batch_size):
                jpeg_filename = self.files[self.i]
                batch.append(jpeg_filename)
                labels.append(label)
                label = label + 1
                self.i = (self.i + 1) % self.n
            labels = np.array(labels).astype('int32')
            return batch, labels

    # Mode 0
    external_input_source = ExternalInputIteratorMode0(batch_size)

    # Create the pipeline
    external_source_pipeline_mode0 = Pipeline(batch_size=batch_size, num_threads=1, device_id=0, prefetch_queue_depth=4,
                                              seed=1, rocal_cpu=True if device == "cpu" else False, tensor_layout=types.NCHW)

    with external_source_pipeline_mode0:
        jpegs, _ = fn.external_source(
            source=external_input_source, mode=types.EXTSOURCE_FNAME)
        output = fn.resize(jpegs, resize_width=300, resize_height=300,
                           output_layout=types.NCHW, output_dtype=types.UINT8)
        external_source_pipeline_mode0.set_outputs(output)

    # build the external_source_pipeline_mode0
    external_source_pipeline_mode0.build()
    # Index starting from 0
    cnt = 0
    # Dataloader
    data_loader = ROCALClassificationIterator(
        external_source_pipeline_mode0, device=device)
    for i, output_list in enumerate(data_loader, 0):
        print("**************MODE 0*******************")
        print("**************", i, "*******************")
        print("**************starts*******************")
        print("\nImages:\n", output_list)
        print("**************ends*******************")
        print("**************", i, "*******************")
        for img in output_list[0][0]:
            cnt = cnt + 1
            image_dump(img, cnt, mode=0)

    ##################### MODE 0 #########################
    
    ##################### MODE 1 #########################
        # Define the Data Source for all image samples
    class ExternalInputIteratorMode1(object):
        def __init__(self, batch_size):
            self.images_dir = data_dir
            self.batch_size = batch_size
            self.files = []
            import os
            self.file_patterns = ['*.jpg', '*.jpeg', '*.JPG', '*.JPEG']
            for pattern in self.file_patterns:
                self.files.extend(glob.glob(os.path.join(self.images_dir, pattern)))

        def __iter__(self):
            self.i = 0
            self.n = len(self.files)
            self.maxWidth = None
            self.maxHeight = None
            return self

        def __next__(self):
            batch = []
            labels = []
            srcsize_height = []
            label = 1
            for x in range(self.batch_size):
                jpeg_filename = self.files[self.i]
                f = open(jpeg_filename, 'rb')
                numpy_buffer = np.frombuffer(f.read(), dtype=np.uint8)
                batch.append(numpy_buffer)
                srcsize_height.append(len(numpy_buffer))
                labels.append(label)
                label = label + 1
                self.i = (self.i + 1) % self.n
            labels = np.array(labels).astype('int32')
            return (batch, labels, srcsize_height)

# Mode 1
    eii_1 = ExternalInputIteratorMode1(batch_size)

    # Create the pipeline
    external_source_pipeline_mode1 = Pipeline(batch_size=batch_size, num_threads=1, device_id=0, prefetch_queue_depth=4,
                                              seed=1, rocal_cpu=True if device == "cpu" else False, tensor_layout=types.NCHW)

    with external_source_pipeline_mode1:
        jpegs, _ = fn.external_source(
            source=eii_1, mode=types.EXTSOURCE_RAW_COMPRESSED, max_width=2000, max_height=2000)
        output = fn.resize(jpegs, resize_width=2000, resize_height=2000,
                           output_layout=types.NCHW, output_dtype=types.UINT8)
        external_source_pipeline_mode1.set_outputs(output)

    # build the external_source_pipeline_mode1
    external_source_pipeline_mode1.build()
    # Index starting from 0
    cnt = 0
    # Dataloader
    data_loader = ROCALClassificationIterator(
        external_source_pipeline_mode1, device=device)
    for i, output_list in enumerate(data_loader, 0):
        print("**************MODE 1*******************")
        print("**************", i, "*******************")
        print("**************starts*******************")
        print("\nImages:\n", output_list)
        print("**************ends*******************")
        print("**************", i, "*******************")
        for img in output_list[0][0]:
            cnt = cnt + 1
            image_dump(img, cnt, mode=1)
    ##################### MODE 1 #########################
    
    ##################### MODE 2 #########################
        # Define the Data Source for all image samples

    class ExternalInputIteratorMode2(object):
        def __init__(self, batch_size):
            self.images_dir = data_dir
            self.batch_size = batch_size
            self.files = []
            self.maxHeight = self.maxWidth = 0
            import os
            self.file_patterns = ['*.jpg', '*.jpeg', '*.JPG', '*.JPEG']
            for pattern in self.file_patterns:
                self.files.extend(glob.glob(os.path.join(self.images_dir, pattern)))
            shuffle(self.files)
            self.i = 0
            self.n = len(self.files)

            for x in range(self.n):
                jpeg_filename = self.files[x]
                label = 1
                image = cv2.imread(jpeg_filename, cv2.IMREAD_COLOR)
                # Check if the image was loaded successfully
                if image is None:
                    print("Error: Failed to load the image.")
                else:
                    # Get the height and width of the image
                    height, width = image.shape[:2]
                self.maxHeight = height if height > self.maxHeight else self.maxHeight
                self.maxWidth = width if width > self.maxWidth else self.maxWidth

        def __iter__(self):
            return self

        def __next__(self):
            batch = []
            batch_of_numpy = []
            labels = []
            label = 1
            roi_height = []
            roi_width = []
            self.out_image = np.zeros(
                (self.batch_size, self.maxHeight, self.maxWidth, 3), dtype="uint8")
            for x in range(self.batch_size):
                jpeg_filename = self.files[self.i]
                image = cv2.imread(jpeg_filename, cv2.IMREAD_COLOR)
                # Check if the image was loaded successfully
                if image is None:
                    print("Error: Failed to load the image.")
                else:
                    # Get the height and width of the image
                    height, width = image.shape[:2]
                batch.append(np.asarray(image))
                roi_height.append(height)
                roi_width.append(width)
                self.out_image[x][:roi_height[x], :roi_width[x], :] = batch[x]
                batch_of_numpy.append(self.out_image[x])
                labels.append(label)
                label = label + 1
                self.i = (self.i + 1) % self.n
            labels = np.array(labels).astype('int32')
            return (batch_of_numpy, labels, roi_height, roi_width, self.maxHeight, self.maxWidth)


# Mode 2
    eii_2 = ExternalInputIteratorMode2(batch_size)

    # Create the pipeline
    external_source_pipeline_mode2 = Pipeline(batch_size=batch_size, num_threads=1, device_id=0, prefetch_queue_depth=4,
                                              seed=1, rocal_cpu=True if device == "cpu" else False, tensor_layout=types.NCHW)

    with external_source_pipeline_mode2:
        jpegs, _ = fn.external_source(source=eii_2, mode=types.EXTSOURCE_RAW_UNCOMPRESSED,
                                      max_width=eii_2.maxWidth, max_height=eii_2.maxHeight)
        output = fn.resize(jpegs, resize_width=300, resize_height=300,
                           output_layout=types.NCHW, output_dtype=types.UINT8)
        external_source_pipeline_mode2.set_outputs(output)

    # build the external_source_pipeline_mode2
    external_source_pipeline_mode2.build()
    # Index starting from 0
    cnt = 0
    # Dataloader
    data_loader = ROCALClassificationIterator(
        external_source_pipeline_mode2, device=device)
    for i, output_list in enumerate(data_loader, 0):
        print("**************MODE 2*******************")
        print("**************", i, "*******************")
        print("**************starts*******************")
        print("\nImages:\n", output_list)
        print("**************ends*******************")
        print("**************", i, "*******************")
        for img in output_list[0][0]:
            cnt = cnt+1
            image_dump(img, cnt, mode=2)
    ##################### MODE 2 #########################
if __name__ == '__main__':
    main()



# Copyright (c) 2024 - 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from PIL import Image
import os
import sys
import datetime
import logging
import ast


def compare_pixels(img1, img2, aug_name, width, height, image_offset=0):
    pixel_difference = [0, 0, 0, 0, 0, 0]
    if "rgb" in aug_name:
        pixels1 = img1.load()
        pixels2 = img2.load()
        channel = 3
    else:
        pixels1 = img1.convert("L").load()
        pixels2 = img2.convert("L").load()
        channel = 1
    total_valid_pixel_count = width * height * channel
    for wt in range(width):
        for ht in range(height):
            ht = ht + image_offset
            if pixels1[wt, ht] != pixels2[wt, ht]:
                if channel == 1:
                    diff_val = abs(pixels1[wt, ht] - pixels2[wt, ht])
                    diff_val = min(diff_val, 5)
                    pixel_difference[diff_val] += 1
                else:
                    for ch in range(channel):
                        diff_val = abs(
                            pixels1[wt, ht][ch] - pixels2[wt, ht][ch])
                        diff_val = min(diff_val, 5)
                        pixel_difference[diff_val] += 1
            else:
                pixel_difference[0] += channel
    return pixel_difference, total_valid_pixel_count

def compare_file_with_list(filename, reference_list):
    with open(filename, 'r') as file:
        file_contents = file.read().strip()
    file_contents = ast.literal_eval(file_contents)
    if len(file_contents) != len(reference_list):
        print("File content and existing list have different lengths.")
        return -1  # Return a negative value to indicate a mismatch in lengths
    mismatch_count = 0
    for file_inner_list, list_inner_list in zip(file_contents, reference_list):
        if file_inner_list != list_inner_list:
            mismatch_count += 1
    return mismatch_count

def main():
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S_")
    handlers = [
        logging.FileHandler("./rocal_unittest_log_file_" + timestamp + ".log"),
        logging.StreamHandler(),
    ]
    logging.basicConfig(level=logging.INFO, handlers=handlers)
    if len(sys.argv) < 3:
        print("Please pass ref_output_folder_path rocal_ouput_folder_path")
        logging.error(
            "Please pass ref_output_folder_path rocal_ouput_folder_path")
        exit(0)
    # Open the two images
    ref_output_path = sys.argv[1]
    rocal_output_path = sys.argv[2]

    if not (os.path.exists(ref_output_path) and os.path.exists(rocal_output_path)):
        logging.error("Path does not Exists")
        exit()
    total_case_count = 0
    passed_case_count = 0
    failed_case_count = 0
    failed_case_list = []
    golden_output_dir_list = os.listdir(ref_output_path)
    rocal_output_dir_list = os.listdir(rocal_output_path)
    randomized_augmentation = ["Snow", "Rain", "Jitter", "SNPNoise"]
    spl_case_meta_data = ["OneHot"]
    golden_file_path = ""
    for aug_name in rocal_output_dir_list:
        temp = aug_name.split(".")
        file_name_split = temp[0].split("_")
        if len(file_name_split) > 3:
            file_name_split.pop()
            golden_file_path = "_".join(file_name_split) + ".png"
        else:
            golden_file_path = aug_name

        # For randomized augmentation
        if file_name_split[0] in randomized_augmentation:
            total_case_count = total_case_count + 1
            augmentation_name = aug_name.split(".")[0]
            logging.info("Running %s", augmentation_name)
            passed_case_count = passed_case_count + 1
            logging.info("PASSED ")
        elif file_name_split[0] in spl_case_meta_data:
            total_case_count = total_case_count + 1
            reference_list = [[1, 0], [1, 0]]
            rocal_file_path = rocal_output_path + aug_name
            if os.path.exists(rocal_file_path):
                logging.info("Running %s ", aug_name.split(".")[0])
            total_mismatches = compare_file_with_list(rocal_file_path, reference_list)
            if total_mismatches == 0:
                passed_case_count = passed_case_count + 1
                logging.info("PASSED")
            else:
                failed_case_list.append(file_name_split[0])
                failed_case_count = failed_case_count + 1
                logging.info("FAILED")
                logging.info("Printing total mismatches in one hot encode %s", total_mismatches)
        elif golden_file_path in golden_output_dir_list:
            total_case_count = total_case_count + 1
            ref_file_path = ref_output_path + golden_file_path
            rocal_file_path = rocal_output_path + aug_name
            if os.path.exists(rocal_file_path) and os.path.exists(ref_file_path):
                logging.info("Running %s ", aug_name.split(".")[0])
                img1 = Image.open(ref_file_path)
                img2 = Image.open(rocal_file_path)

                # Check if the images have the same dimensions
                if img1.size != img2.size:
                    logging.info(
                        "Golden output and augmentation outputs are having different sizes. Exiting!")
                    exit()

                # Compare the pixel values for each image
                pixel_diff = None
                total_count = 0
                if "larger" in aug_name:
                    resize_width = 400
                    resize_height = 400
                    image_offset = 400
                    pixel_diff, total_count = compare_pixels(
                        img1, img2, aug_name, resize_width, resize_height)
                    pixel_diff2, total_count2 = compare_pixels(
                        img1, img2, aug_name, resize_width, resize_height, image_offset)
                    pixel_diff = [x + y for x,
                                  y in zip(pixel_diff, pixel_diff2)]
                    total_count = total_count + total_count2
                elif "smaller" in aug_name:
                    resize_width = 400
                    resize_height = 400
                    image_offset = 2400
                    pixel_diff, total_count = compare_pixels(
                        img1, img2, aug_name, resize_width, resize_height)
                    pixel_diff2, total_count2 = compare_pixels(
                        img1, img2, aug_name, resize_width, resize_height, image_offset)
                    pixel_diff = [x + y for x,
                                  y in zip(pixel_diff, pixel_diff2)]
                    total_count = total_count + total_count2
                else:
                    pixel_diff, total_count = compare_pixels(
                        img1, img2, aug_name, img1.size[0], img1.size[1])
                total_pixel_diff = 0
                for pix_diff in range(1, 6):
                    total_pixel_diff += pixel_diff[pix_diff]
                mismatch_percentage = round(
                    (total_pixel_diff / total_count) * 100, 2)
                if ((total_pixel_diff == 0) or (mismatch_percentage < 5.0 and pixel_diff[1] == total_pixel_diff) or       # Ignore test cases with single pixel differences less than 5% of total pixel count
                        (mismatch_percentage < 0.5 and ("Blend" in aug_name or "Rotate" in aug_name) and "hip" in aug_name)):  # Ignore mismatch in rotate augmentation less than 0.5% of total pixel count
                    passed_case_count = passed_case_count + 1
                    logging.info("PASSED")
                else:
                    failed_case_list.append(golden_file_path)
                    failed_case_count = failed_case_count + 1
                    logging.info("FAILED")
                    logging.info("Printing pixel mismatch %s", pixel_diff)
                    logging.info("Mismatach percentage %0.2f",
                                 mismatch_percentage)
                    for pix_diff in range(1, 6):
                        logging.info("Percentage of %d pixel mismatch %0.2f", pix_diff, round(
                            (pixel_diff[pix_diff] / total_pixel_diff) * 100, 2))
            else:
                logging.info(
                    "Skipping the testcase as file not found %s", rocal_file_path)
        else:
            logging.info("File not found in ref_output_folder %s",
                         golden_file_path)
    if len(failed_case_list) != 0:
        logging.info("Failing cases: {}".format(", ".join(failed_case_list)))
    logging.info(
        "Total case passed --> {} / {} ".format(passed_case_count, total_case_count))
    logging.info(
        "Total case failed --> {} / {} ".format(failed_case_count, total_case_count))


if __name__ == "__main__":
    main()
